import sys
from typing import (
    TypeVar,
    Any,
    Optional,
    List,
    Sequence,
    Tuple,
    Iterable,
    Generic,
    Iterator,
    Callable,
    Hashable,
    Dict,
    Set,
    Union,
)
from robocorp_ls_core.protocols import (
    Sentinel,
    IMonitor,
    IDocument,
    IWorkspace,
    IConfig,
    IDocumentSelection,
    ITestInfoFromSymbolsCacheTypedDict,
)
from robocorp_ls_core.constants import NULL
from robocorp_ls_core.protocols import TypedDict
import enum
from robocorp_ls_core.lsp import (
    LocationTypedDict,
    RangeTypedDict,
    MarkupContentTypedDict,
    CompletionItemTypedDict,
    LSPMessages,
)
import typing
from robocorp_ls_core.ordered_set import OrderedSet
from contextlib import contextmanager

if sys.version_info[:2] < (3, 8):

    class Protocol(object):
        pass


else:
    from typing import Protocol

T = TypeVar("T")
Y = TypeVar("Y", covariant=True)

# We don't want to import robot in this case (just do it when type-checking).
class IRobotToken(Protocol):
    SETTING_HEADER: str
    VARIABLE_HEADER: str
    TESTCASE_HEADER: str
    KEYWORD_HEADER: str
    COMMENT_HEADER: str

    TESTCASE_NAME: str
    KEYWORD_NAME: str

    DOCUMENTATION: str
    SUITE_SETUP: str
    SUITE_TEARDOWN: str
    METADATA: str
    TEST_SETUP: str
    TEST_TEARDOWN: str
    TEST_TEMPLATE: str
    TEST_TIMEOUT: str
    FORCE_TAGS: str
    DEFAULT_TAGS: str
    LIBRARY: str
    RESOURCE: str
    VARIABLES: str
    SETUP: str
    TEARDOWN: str
    TEMPLATE: str
    TIMEOUT: str
    TAGS: str
    ARGUMENTS: str
    # Use ´RETURN_SETTING` type instead of `RETURN`. `[Return]` is deprecated and
    # `RETURN` type will be used with `RETURN` statement in the future.
    RETURN: str
    RETURN_SETTING: str

    NAME: str
    VARIABLE: str
    ARGUMENT: str
    ASSIGN: str
    KEYWORD: str
    WITH_NAME: str
    FOR: str
    FOR_SEPARATOR: str
    END: str
    IF: str
    INLINE_IF: str
    ELSE_IF: str
    ELSE: str
    TRY: str
    EXCEPT: str
    FINALLY: str
    AS: str
    WHILE: str
    RETURN_STATEMENT: str
    CONTINUE: str
    BREAK: str

    SEPARATOR: str
    COMMENT: str
    CONTINUATION: str
    EOL: str
    EOS: str

    ERROR: str
    FATAL_ERROR: str

    type: str
    value: str
    lineno: int  # 1-based
    col_offset: int  # 0-based
    error: Any

    @property
    def end_col_offset(self) -> int:  # 0-based
        pass

    def tokenize_variables(self) -> Iterator["IRobotToken"]:
        pass


class IRobotVariableMatch(Protocol):
    string: str
    identifier: str
    base: Optional[str]
    items: Tuple[str, ...]
    start: int
    end: int

    @property
    def name(self) -> str:
        pass

    @property
    def before(self) -> str:
        pass

    @property
    def match(self) -> str:
        pass

    @property
    def after(self) -> str:
        pass


class INode(Protocol):
    type: str
    lineno: int
    end_lineno: int
    col_offset: int
    end_col_offset: int
    tokens: List[IRobotToken]

    def get_token(self, name: str) -> Optional[IRobotToken]:
        pass

    def get_tokens(self, name: str) -> List[IRobotToken]:
        pass


class ILibraryImportNode(INode, Protocol):
    name: str
    alias: Optional[str]
    args: Optional[Sequence[str]]


class IKeywordNode(INode, Protocol):
    name: str


class IResourceImportNode(INode, Protocol):
    name: str


class IVariableImportNode(INode, Protocol):
    name: str


class NodeInfo(Generic[Y]):
    stack: Tuple[INode, ...]
    node: Y

    __slots__ = ["stack", "node"]

    def __init__(self, stack, node):
        self.stack = stack
        self.node = node

    def __str__(self):
        return f"NodeInfo({self.node.__class__.__name__})"

    __repr__ = __str__


class TokenInfo:

    __slots__ = ["stack", "node", "token"]

    def __init__(self, stack: Tuple[INode, ...], node: INode, token: IRobotToken):
        self.stack = stack
        self.node = node
        self.token = token

    def __str__(self):
        return f"TokenInfo({self.token.value} -- in: {self.node.__class__.__name__})"

    __repr__ = __str__


class AdditionalVarInfo:

    CONTEXT_UNDEFINED = 0
    CONTEXT_EXPRESSION = 1

    __slots__ = ["var_identifier", "context", "extended_part"]

    def __init__(
        self,
        var_identifier: str = "",
        context: int = CONTEXT_UNDEFINED,
        extended_part: str = "",
    ):
        """
        :param var_identifier: One of: $,@,%
        """

        self.var_identifier = var_identifier
        self.context = context
        self.extended_part = extended_part

    def copy(self, **kwargs):
        new_kwargs = {
            "var_identifier": self.var_identifier,
            "context": self.context,
            "extended_part": self.extended_part,
        }
        new_kwargs.update(kwargs)
        return AdditionalVarInfo(**new_kwargs)

    def __str__(self):
        info = [f"AdditionalVarInfo({self.var_identifier}"]
        if self.context:
            info.append(f" -- ctx: {self.context}")
        if self.extended_part:
            info.append(f" -- extended: {self.extended_part}")
        info.append(")")
        return "".join(info)

    __repr__ = __str__


class VarTokenInfo:

    __slots__ = ["stack", "node", "token", "var_info"]

    def __init__(
        self,
        stack: Tuple[INode, ...],
        node: INode,
        token: IRobotToken,
        var_info: AdditionalVarInfo,
    ):
        self.stack = stack
        self.node = node
        self.token = token
        self.var_info = var_info

    def __str__(self):
        return f"VarTokenInfo({self.token.value} (in {self.node.__class__.__name__}) - {self.var_info})"

    __repr__ = __str__


class KeywordUsageInfo:
    __slots__ = ["stack", "node", "token", "name", "_is_argument_usage", "prefix"]

    def __init__(
        self,
        stack: Tuple[INode, ...],
        node: INode,
        token: IRobotToken,
        name: str,
        is_argument_usage: bool = False,
        prefix: str = "",
    ):
        self.stack = stack
        self.node = node
        self.token = token  # This is actually the keyword name token.
        self.name = name
        self._is_argument_usage = is_argument_usage
        self.prefix = prefix

    def __repr__(self):
        if self._is_argument_usage:
            return f"KeywordUsageInfo({self.name} - {self.node.__class__.__name__} (argument usage))"
        else:
            return f"KeywordUsageInfo({self.name} - {self.node.__class__.__name__})"

    __str__ = __repr__


class IKeywordArg(Protocol):
    @property
    def original_arg(self) -> str:
        pass

    @property
    def arg_name(self) -> str:
        pass

    @property
    def is_keyword_arg(self) -> bool:
        pass

    @property
    def is_star_arg(self) -> bool:
        pass

    def is_arg_type_set(self) -> bool:
        pass

    @property
    def arg_type(self) -> Optional[str]:
        pass

    def is_default_value_set(self) -> bool:
        pass

    @property
    def default_value(self) -> Optional[str]:
        pass


class ILibraryDoc(Protocol):
    filename: str
    name: str
    source: str
    symbols_cache: Optional["ISymbolsCache"]
    inits: list
    doc_format: str
    keywords: List["IKeywordDoc"]


class ILibraryDocConversions(ILibraryDoc):
    """
    Note: these are actually part of the basic library doc but we
    put it in a different interface because clients usually shouldn't
    use it (it's controlled by the libspec manager).
    """

    def convert_docs_to_html(self):
        pass

    def convert_docs_to_markdown(self):
        pass


class IKeywordDoc(Protocol):
    name: str
    tags: Tuple[str, ...]
    lineno: int
    doc: str

    @property
    def args(self) -> Tuple[IKeywordArg, ...]:
        pass

    @property
    def libdoc(self) -> ILibraryDoc:
        pass

    @property
    def deprecated(self) -> bool:
        pass

    @property
    def source(self) -> str:
        pass

    @property
    def doc_format(self) -> str:
        pass

    def to_dictionary(self) -> dict:
        pass


class ILibraryDocOrError(Protocol):
    library_doc: Optional[ILibraryDoc]
    error: Optional[str]


class IRobotDocument(IDocument, Protocol):
    def get_type(self) -> str:
        pass

    def get_ast(self) -> Any:
        pass

    def get_python_ast(self) -> Optional[Any]:
        pass

    def get_yaml_contents(self) -> Optional[Any]:
        pass

    symbols_cache: Optional["ISymbolsCache"]


class ISymbolsJsonListEntry(TypedDict):
    name: str
    kind: int  # SymbolKind
    location: LocationTypedDict
    containerName: str


class ISymbolKeywordInfo(Protocol):
    name: str

    def get_documentation(self) -> MarkupContentTypedDict:
        """
        Note: It should be computed on demand (and can be slow).
        """


class ISymbolsCache(Protocol):
    def get_uri(self) -> Optional[str]:
        """
        If we're referencing a library (and have the symbols from a libspec),
        the uri may be None.
        """

    def has_keyword_usage(self, normalized_keyword_name: str) -> bool:
        pass

    def has_global_variable_definition(self, normalized_variable_name: str) -> bool:
        pass

    def has_variable_reference(self, normalized_variable_name: str) -> bool:
        pass

    def get_json_list(self) -> List[ISymbolsJsonListEntry]:
        pass

    def get_library_info(self) -> Optional[ILibraryDoc]:
        pass

    def get_doc(self) -> Optional[IRobotDocument]:
        pass

    def get_test_info(self) -> Optional[List[ITestInfoFromSymbolsCacheTypedDict]]:
        pass

    def iter_keyword_info(self) -> Iterator[ISymbolKeywordInfo]:
        pass


class ICompletionContextWorkspaceCaches(Protocol):
    cache_hits: int

    def on_file_changed(self, filename: str):
        pass

    def on_updated_document(self, uri: str, document: Optional[IRobotDocument]):
        pass

    def clear_caches(self):
        pass

    def dispose(self):
        pass

    def get_cached_dependency_graph(
        self, cache_key: Hashable
    ) -> Optional["ICompletionContextDependencyGraph"]:
        pass

    @contextmanager
    def invalidation_tracker(self):
        """
        Note that it's possible that changes happen in-flight. This means that
        we must track changes while the dependency graph is being calculated.

        So, one must do something as:

        with caches.invalidation_tracker() as invalidation_tracker:
            ... compute dependency graph
            caches.cache_dependency_graph(cache_key, dependency_graph, invalidation_tracker)
        """

    def cache_dependency_graph(
        self,
        cache_key: Hashable,
        dependency_graph: "ICompletionContextDependencyGraph",
        invalidation_tracker,
    ) -> None:
        pass


class IRobotWorkspace(IWorkspace, Protocol):
    completion_context_workspace_caches: ICompletionContextWorkspaceCaches
    libspec_manager: Any

    def iter_all_doc_uris_in_workspace(
        self, extensions: Tuple[str, ...]
    ) -> Iterable[str]:
        pass


class IKeywordFound(Protocol):
    """
    :ivar completion_context:
        This may be a new completion context, created when a new document is
        being analyzed (the keyword was created for that completion context).
        For libraries the initial completion context is passed.
    :ivar source:
        Source where the keyword was found.
    :ivar lineno:
        Line where it was found (0-based).
    """

    @property
    def keyword_name(self) -> str:
        pass

    @property
    def keyword_ast(self) -> Optional[INode]:
        """
        Only available when we do have a keyword AST (i.e.: not for library
        keywords).
        """

    @property
    def keyword_args(self) -> Sequence[IKeywordArg]:
        pass

    def is_deprecated(self) -> bool:
        pass

    def compute_docs_with_signature(self) -> MarkupContentTypedDict:
        pass

    def compute_docs_without_signature(self) -> MarkupContentTypedDict:
        pass

    completion_context: Optional["ICompletionContext"]

    completion_item_kind: int = -1

    @property
    def source(self) -> str:
        """
        Provides the filesystem location where the keyword was found.
        """

    @property
    def lineno(self) -> int:
        pass

    @property
    def end_lineno(self) -> int:
        pass

    @property
    def col_offset(self) -> int:
        pass

    @property
    def end_col_offset(self) -> int:
        pass

    @property
    def library_name(self) -> Optional[str]:
        # If it's a library, this is the name of the library.
        pass

    @property
    def resource_name(self) -> Optional[str]:
        # If it's a resource, this is the basename of the resource without the extension.
        pass

    @property
    def library_alias(self) -> Optional[str]:
        pass

    # These are added if possible if there's some range to include the
    # full scope of the keyword. It should always include
    # the lineno/end_lineno range (so, it's a superset).

    @property
    def scope_lineno(self) -> Optional[int]:
        pass

    @property
    def scope_end_lineno(self) -> Optional[int]:
        pass

    @property
    def scope_col_offset(self) -> Optional[int]:
        pass

    @property
    def scope_end_col_offset(self) -> Optional[int]:
        pass


class IKeywordCollector(Protocol):
    def accepts(self, keyword_name: str) -> bool:
        """
        :param keyword_name:
            The name of the keyword to be accepted or not.
        :return bool:
            If the return is True, on_keyword(...) is called (otherwise it's not
            called).
        """

    def on_keyword(self, keyword_found: IKeywordFound):
        """
        :param IKeywordFound keyword_found:
        """

    def on_resolved_library(
        self,
        completion_context: "ICompletionContext",
        library_node: Optional[INode],
        library_doc: "ILibraryDoc",
    ):
        pass

    def on_unresolved_library(
        self,
        completion_context: "ICompletionContext",
        library_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
        resolved_name: str,
    ):
        pass

    def on_unresolved_resource(
        self,
        completion_context: "ICompletionContext",
        resource_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
        resolved_name: str,
    ):
        pass


class AbstractKeywordCollector:
    def on_resolved_library(
        self,
        completion_context: "ICompletionContext",
        library_node,
        library_doc: ILibraryDoc,
    ):
        pass

    def on_unresolved_library(
        self,
        completion_context: "ICompletionContext",
        library_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
        resolved_name: str,
    ):
        pass

    def on_unresolved_resource(
        self,
        completion_context: "ICompletionContext",
        resource_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
        resolved_name: str,
    ):
        pass


class IDefinition(Protocol):

    keyword_name: str = ""  # Can be empty if it's not found as a keyword.

    # Note: Could be None (i.e.: we found it in a library spec file which doesn't have the source).
    source: str = ""

    # Note: if we found it in a library spec file which doesn't have the lineno, it should be 0
    lineno: int = 0

    # Note: if we found it in a library spec file which doesn't have the lineno, it should be 0
    end_lineno: int = 0

    col_offset: int = 0

    end_col_offset: int = 0

    # These are added if possible if there's some range to include the
    # full scope (of a keyword, test, etc). It should always include
    # the lineno/end_lineno range (so, it's a superset).
    scope_lineno: Optional[int] = None
    scope_end_lineno: Optional[int] = None
    scope_col_offset: Optional[int] = None
    scope_end_col_offset: Optional[int] = None

    def hover_docs(self) -> MarkupContentTypedDict:
        pass


class IKeywordDefinition(IDefinition, Protocol):

    keyword_found: IKeywordFound


class IVariableDefinition(IDefinition, Protocol):

    variable_found: "IVariableFound"


def cast_to_keyword_definition(definition: IDefinition) -> Optional[IKeywordDefinition]:
    if hasattr(definition, "keyword_found"):
        return typing.cast(IKeywordDefinition, definition)
    return None


def cast_to_variable_definition(
    definition: IDefinition,
) -> Optional[IVariableDefinition]:
    if hasattr(definition, "variable_found"):
        return typing.cast(IVariableDefinition, definition)
    return None


class IBaseCompletionContext(Protocol):
    @property
    def monitor(self) -> Optional[IMonitor]:
        pass

    @property
    def workspace(self) -> IRobotWorkspace:
        pass

    @property
    def config(self) -> Optional[IConfig]:
        pass

    def check_cancelled(self) -> None:
        pass


class CompletionType(enum.Enum):
    regular = 1
    shell = 2


class LibraryDependencyInfo:
    def __init__(
        self,
        name: str,
        alias: Optional[str],
        builtin: bool,
        args: Optional[str],
        node: Optional[ILibraryImportNode],
    ):
        """
        :param builtin:
            Note that builtin should only be set == True if it's actually known
            that it's a builtin, otherwise it should be set to False (in which
            case it's computed internally if it's builtin or not).
        """
        self.name = name
        self.alias = alias
        self.builtin = builtin
        self.args = args
        self.node = node

    def to_dict(self):
        ret = {
            "name": self.name,
        }
        if self.alias:
            ret["alias"] = self.alias
        if self.builtin:
            ret["builtin"] = self.builtin
        if self.args:
            ret["args"] = self.args
        return ret


class ISymbolsCacheReverseIndex(Protocol):
    def get_global_variable_uri_definitions(
        self, normalized_var_name: str
    ) -> Optional[Set[str]]:
        pass

    def has_global_variable(self, normalized_var_name: str) -> bool:
        pass


class ICompletionContextDependencyGraph(Protocol):
    def add_library_infos(
        self,
        doc_uri: str,
        library_infos: OrderedSet[LibraryDependencyInfo],
    ):
        pass

    def add_resource_infos(
        self,
        doc_uri: str,
        resource_imports_as_docs: Sequence[
            Tuple[IResourceImportNode, Optional[IRobotDocument]]
        ],
    ):
        pass

    def add_variable_infos(
        self,
        doc_uri: str,
        new_variable_imports: Sequence[
            Tuple[IVariableImportNode, Optional[IRobotDocument]]
        ],
    ):
        pass

    def get_root_doc(self) -> IRobotDocument:
        pass

    def iter_libraries(self, doc_uri: str) -> Iterator[LibraryDependencyInfo]:
        """
        Provides an iterator(doc_uri, library_dependency_infos)
        """

    def iter_all_libraries(self) -> Iterator[LibraryDependencyInfo]:
        pass

    def iter_resource_imports_with_docs(
        self, doc_uri: str
    ) -> Iterator[Tuple[IResourceImportNode, Optional[IRobotDocument]]]:
        pass

    def iter_all_resource_imports_with_docs(
        self,
    ) -> Iterator[Tuple[IResourceImportNode, Optional[IRobotDocument]]]:
        pass

    def iter_variable_imports_as_docs(
        self, doc_uri: str
    ) -> Iterator[Tuple[IVariableImportNode, Optional[IRobotDocument]]]:
        pass

    def iter_all_variable_imports_as_docs(
        self,
    ) -> Iterator[Tuple[IVariableImportNode, Optional[IRobotDocument]]]:
        pass

    def to_dict(self) -> dict:
        pass

    def do_invalidate_on_uri_change(self, uri: str) -> bool:
        pass


class IVariablesFromArgumentsFileLoader(Protocol):
    def get_variables(self) -> Tuple["IVariableFound", ...]:
        pass


class ILocalizationInfo(Protocol):
    def __init__(self, language_codes: Union[Tuple[str, ...], str]):
        pass

    @property
    def language_codes(self) -> Tuple[str, ...]:
        pass

    def iter_bdd_prefixes_on_read(self) -> Iterator[str]:
        """
        Note that we specify the reason for iterating because for instance, when
        writing code we could want just the completions for the specified
        language in the file and while reading (i.e.: analyzing) we'd want it
        for all languages.
        """

    def iter_languages_on_write(
        self,
    ) -> Iterator[Any]:  # Actually Iterator[robot.api.Language]
        """
        Provides the languages used when writing a doc (i.e.: completions, ...).
        """


class ICompletionContext(Protocol):
    tracing: bool

    def __init__(
        self,
        doc,
        line=Sentinel.SENTINEL,
        col=Sentinel.SENTINEL,
        workspace=None,
        config=None,
        memo=None,
        monitor: IMonitor = NULL,
        variables_from_arguments_files_loader: Sequence[
            IVariablesFromArgumentsFileLoader
        ] = (),
    ) -> None:
        pass

    def resolve_completion_item(
        self, data, completion_item: CompletionItemTypedDict, monaco: bool = False
    ) -> None:
        pass

    @property
    def lsp_messages(
        self,
    ) -> Optional[LSPMessages]:
        pass

    @property
    def variables_from_arguments_files_loader(
        self,
    ) -> Sequence[IVariablesFromArgumentsFileLoader]:
        pass

    @property
    def type(self) -> CompletionType:
        pass

    @property
    def monitor(self) -> IMonitor:
        pass

    def check_cancelled(self):
        pass

    def create_copy_with_selection(self, line: int, col: int) -> "ICompletionContext":
        pass

    def create_copy(self, doc: IRobotDocument) -> "ICompletionContext":
        pass

    def create_copy_with_config(self, config: IConfig) -> "ICompletionContext":
        pass

    def create_copy_doc_line_col(
        self, doc: IRobotDocument, line: int, col: int
    ) -> "ICompletionContext":
        pass

    @property
    def original_doc(self) -> IRobotDocument:
        pass

    @property
    def original_sel(self) -> Any:
        pass

    @property
    def doc(self) -> IRobotDocument:
        pass

    @property
    def sel(self) -> IDocumentSelection:
        pass

    @property
    def memo(self) -> Any:
        pass

    @property
    def config(self) -> Optional[IConfig]:
        pass

    @property
    def workspace(self) -> IRobotWorkspace:
        pass

    def get_type(self) -> Any:
        pass

    def get_ast(self) -> Any:
        pass

    def get_ast_current_section(self) -> Optional[INode]:
        """
        :rtype: robot.parsing.model.blocks.Section|NoneType
        """

    def get_current_section_name(self) -> Optional[str]:
        pass

    def get_current_token(self) -> Optional[TokenInfo]:
        pass

    def get_all_variables(self) -> Tuple[NodeInfo, ...]:
        pass

    def get_doc_normalized_var_name_to_var_found(self) -> Dict[str, "IVariableFound"]:
        pass

    def get_settings_normalized_var_name_to_var_found(
        self,
    ) -> Dict[str, "IVariableFound"]:
        pass

    def get_builtins_normalized_var_name_to_var_found(
        self, resolved
    ) -> Dict[str, "IVariableFound"]:
        pass

    def get_arguments_files_normalized_var_name_to_var_found(
        self,
    ) -> Dict[str, "IVariableFound"]:
        pass

    def get_current_variable(self, section=None) -> Optional[VarTokenInfo]:
        """
        Provides the current variable token. Note that it won't include '{' nor '}'.
        """

    def get_resource_import_as_doc(
        self, resource_import: INode, check_as_module: bool = False
    ) -> Optional[IRobotDocument]:
        pass

    def get_variable_imports(self) -> Tuple[INode, ...]:
        pass

    def get_variable_import_as_doc(self, variables_import) -> Optional[IRobotDocument]:
        pass

    def get_current_keyword_definition(self) -> Optional[IKeywordDefinition]:
        pass

    def get_resource_imports(
        self,
    ) -> Tuple[IResourceImportNode, ...]:
        pass

    def get_resource_imports_as_docs(
        self,
    ) -> Tuple[Tuple[IResourceImportNode, Optional[IRobotDocument]], ...]:
        pass

    def get_resource_inits_as_docs(self) -> Tuple[IRobotDocument, ...]:
        pass

    def get_variable_imports_as_docs(
        self,
    ) -> Tuple[Tuple[IVariableImportNode, Optional[IRobotDocument]], ...]:
        pass

    def get_imported_libraries(self) -> Tuple[ILibraryImportNode, ...]:
        pass

    def token_value_resolving_variables(self, token: IRobotToken) -> str:
        pass

    def token_value_and_unresolved_resolving_variables(
        self, token: IRobotToken
    ) -> Tuple[str, Tuple[Tuple[IRobotToken, str], ...]]:
        pass

    def get_current_keyword_definition_and_usage_info(
        self,
    ) -> Optional[Tuple[IKeywordDefinition, KeywordUsageInfo]]:
        pass

    def get_current_keyword_usage_info(
        self,
    ) -> Optional[KeywordUsageInfo]:
        pass

    def assign_documentation_resolve(
        self,
        completion_item: CompletionItemTypedDict,
        compute_documentation: Callable[[], MarkupContentTypedDict],
    ) -> None:
        pass

    def collect_dependency_graph(self) -> ICompletionContextDependencyGraph:
        pass

    def iter_dependency_and_init_resource_docs(
        self, dependency_graph
    ) -> Iterator[IRobotDocument]:
        pass

    def obtain_symbols_cache_reverse_index(self) -> Optional[ISymbolsCacheReverseIndex]:
        pass

    def get_ast_localization_info(self) -> ILocalizationInfo:
        pass


class VariableKind:
    VARIABLE = "Variable"
    BUILTIN = "Builtin Variable"
    ARGUMENT = "Argument"
    ENV_VARIABLE = "Environment Variable"
    SETTINGS = "Variable (settings)"
    PYTHON = "Variable (python)"
    YAML = "Variable (yaml)"
    ARGUMENTS_FILE = "Arguments file"
    LOCAL_ASSIGN_VARIABLE = "Variable (local assign)"
    LOCAL_SET_VARIABLE = "Variable (local set)"
    TASK_SET_VARIABLE = "Variable (task set)"
    TEST_SET_VARIABLE = "Variable (test set)"
    SUITE_SET_VARIABLE = "Variable (suite set)"
    GLOBAL_SET_VARIABLE = "Variable (global)"
    ENV_SET_VARIABLE = "Variable (environment)"


LOCAL_ASSIGNS_VARIABLE_KIND = {
    VariableKind.ARGUMENT,
    VariableKind.LOCAL_ASSIGN_VARIABLE,
    VariableKind.LOCAL_SET_VARIABLE,
}


class IVariableFound(Protocol):
    """
    :ivar variable_name:
        This is the value that we should use when completing.
        It's the name of the variable without `${}` chars.

    :ivar variable_value:
        The value of the variable -- in general used to show information
        regarding that variable to the user.

    :ivar completion_context:
        This may be a new completion context, created when a new document is
        being analyzed (the variable was created for that completion context).

    :ivar source:
        Source where the variable was found.

    :ivar lineno:
        Line where it was found (0-based).

    :ivar stack:
        The stack where the variable was found (only available if it was
        found in a robot file where the ast is available -- i.e.: settings,
        yaml, python, etc. variables don't have a stack available).
    """

    variable_name: str = ""
    variable_value: str = ""
    variable_kind: str = VariableKind.VARIABLE
    completion_context: Optional[ICompletionContext] = None
    stack: Optional[Tuple[INode, ...]] = None

    @property
    def is_local_variable(self) -> bool:
        pass

    @property
    def source(self) -> str:
        pass

    # Note: line/offsets 0-based.
    @property
    def lineno(self) -> int:
        pass

    @property
    def end_lineno(self) -> int:
        pass

    @property
    def col_offset(self) -> int:
        pass

    @property
    def end_col_offset(self) -> int:
        pass


class IVariablesCollector(Protocol):
    def accepts(self, variable_name: str) -> bool:
        """
        :param variable_name:
            The name of the variable (i.e.: ${some_var}).
        """

    def on_variable(self, variable_found: IVariableFound):
        pass

    def on_unresolved_variable_import(
        self,
        completion_context: "ICompletionContext",
        variable_import_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
        resolved_name: str,
    ):
        pass

    def on_env_variable(self, variable_found: IVariableFound):
        """
        Called for environment variables using the
        'Set Keyword Variable' keyword.

        Note: doesn't call `accepts` first.
        """


class AbstractVariablesCollector:
    def on_env_variable(self, variable_found: IVariableFound):
        pass

    def on_unresolved_variable_import(
        self,
        completion_context: "ICompletionContext",
        variable_import_name: str,
        lineno: int,
        end_lineno: int,
        col_offset: int,
        end_col_offset: int,
        error_msg: Optional[str],
        resolved_name: str,
    ):
        pass


class EvaluatableExpressionTypedDict(TypedDict):
    range: RangeTypedDict
    expression: Optional[str]


class IOnDependencyChanged(Protocol):
    def __call__(self, uri: str):
        pass



import robot.errors


class RobocopFatalError(ValueError):
    pass


class ConfigGeneralError(RobocopFatalError):
    pass


class InvalidExternalCheckerError(RobocopFatalError):
    def __init__(self, path):
        msg = f'Fatal error: Failed to load external rules from file "{path}". Verify if the file exists'
        super().__init__(msg)


class FileError(RobocopFatalError):
    def __init__(self, source):
        msg = f'File "{source}" does not exist'
        super().__init__(msg)


class ArgumentFileNotFoundError(RobocopFatalError):
    def __init__(self, source):
        msg = f'Argument file "{source}" does not exist'
        super().__init__(msg)


class NestedArgumentFileError(RobocopFatalError):
    def __init__(self, source):
        msg = f'Nested argument file in "{source}"'
        super().__init__(msg)


class InvalidArgumentError(RobocopFatalError):
    def __init__(self, msg):
        super().__init__(f"Invalid configuration for Robocop:\n{msg}")


class RuleNotFoundError(RobocopFatalError):
    def __init__(self, rule, checker):
        super().__init__(
            f"{checker.__class__.__name__} checker does not contain rule `{rule}`. "
            f"Available rules: {', '.join(checker.rules.keys())}"
        )


class RuleParamNotFoundError(RobocopFatalError):
    def __init__(self, rule, param, checker):
        super().__init__(
            f"Rule `{rule.name}` in `{checker.__class__.__name__}` checker does not contain `{param}` param. "
            f"Available params:\n    {rule.available_configurables()[1]}"
        )


class RuleParamFailedInitError(RobocopFatalError):
    def __init__(self, param, value, err):
        desc = f"    Parameter info: {param.desc}" if param.desc else ""
        super().__init__(
            f"Failed to configure param `{param.name}` with value `{value}`. Received error `{err}`.\n"
            f"    Parameter type: {param.converter}\n" + desc
        )


class RuleReportsNotFoundError(RobocopFatalError):
    def __init__(self, rule, checker):
        super().__init__(f"{checker.__class__.__name__} checker `reports` attribute contains unknown rule `{rule}`")


class InvalidReportName(ConfigGeneralError):
    def __init__(self, report, reports):
        from robocop.utils import RecommendationFinder

        report_names = sorted(list(reports.keys()) + ["all"])
        similar = RecommendationFinder().find_similar(report, report_names)
        msg = f"Provided report '{report}' does not exist. {similar}"
        super().__init__(msg)


class RobotFrameworkParsingError(Exception):
    def __init__(self):
        msg = (
            "Fatal exception occurred when using Robot Framework parsing module. "
            "Consider updating Robot Framework to recent stable version."
        )
        super().__init__(msg)


def handle_robot_errors(func):
    """
    If the user uses older version of Robot Framework, it many fail while parsing the
    source code due to bug that is already fixed in the more recent version.
    """

    def wrap_errors(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except robot.errors.DataError:
            raise
        except:  # noqa
            raise RobotFrameworkParsingError

    return wrap_errors



from collections import defaultdict
from typing import List, Tuple

from robot.api.parsing import ElseHeader, ElseIfHeader, ModelVisitor, Token
from robot.parsing.model import Statement

try:
    from robot.api.parsing import InlineIfHeader, TryHeader
except ImportError:
    InlineIfHeader, TryHeader = None, None

from robotidy.disablers import skip_if_disabled
from robotidy.exceptions import InvalidParameterValueError
from robotidy.skip import Skip
from robotidy.transformers import Transformer
from robotidy.utils import is_blank_multiline, round_to_four

WHITESPACE_TOKENS = frozenset({Token.SEPARATOR, Token.EOS})


class AlignKeywordsTestsSection(Transformer):

    ENABLED = False
    DEFAULT_WIDTH = 24
    HANDLES_SKIP = frozenset(
        {
            "skip_documentation",
            "skip_return_values",
            "skip_keyword_call",
            "skip_keyword_call_pattern",
            "skip_settings",
            "skip_arguments",
            "skip_setup",
            "skip_teardown",
            "skip_template",
            "skip_timeout",
            "skip_return",
            "skip_tags",
        }
    )

    def __init__(
        self,
        widths: str,
        alignment_type: str,
        handle_too_long: str,
        compact_overflow_limit: int = 2,
        skip: Skip = None,
    ):
        super().__init__(skip)
        self.is_inline = False
        self.indent = 1
        self.handle_too_long = self.parse_handle_too_long(handle_too_long)
        self.fixed_alignment = self.parse_alignment_type(alignment_type)
        self.compact_overflow_limit = compact_overflow_limit
        self.split_too_long = False
        # column widths map - 0: 40, 1: 30
        if widths:
            self.widths = self.parse_widths(widths)
        else:
            self.widths = None
        self.auto_widths = []

    @skip_if_disabled
    def visit_File(self, node):  # noqa
        self.split_too_long = self.is_split_too_long_enabled()
        return self.generic_visit(node)

    def parse_widths(self, widths):
        parsed_widths = dict()
        for index, width in enumerate(widths.split(",")):
            try:
                number = int(width)
                if number < 0:
                    raise ValueError("Should not be a negative number.")
            except ValueError:
                raise InvalidParameterValueError(
                    self.__class__.__name__,
                    "widths",
                    widths,
                    "Widths should be comma separated list of numbers equal or greater than 0.",
                ) from None
            parsed_widths[index] = number
        return parsed_widths

    def parse_handle_too_long(self, value):
        if value not in ("overflow", "compact_overflow", "ignore_line", "ignore_rest"):
            raise InvalidParameterValueError(
                self.__class__.__name__,
                "handle_too_long",
                value,
                "Chose between modes: 'overflow' (align to the next column), 'compact_overflow' (), "
                "'ignore_line' (ignore this line in alignment) or 'ignore_rest' (align to long token and ignore rest)",
            )
        return value

    def parse_alignment_type(self, value):
        if value not in ("fixed", "auto"):
            raise InvalidParameterValueError(
                self.__class__.__name__,
                "alignment_type",
                value,
                "Chose between two modes: 'fixed' (align to fixed width) or "
                "'auto' (align to longest token in column).",
            )
        return value == "fixed"

    def parse_documentation_mode(self, doc_mode):
        if doc_mode not in ("skip", "align_first_col"):
            raise InvalidParameterValueError(
                self.__class__.__name__,
                "doc_mode",
                doc_mode,
                "Chose between two modes: 'skip' (default - do not align documentation) or "
                "'align_first_col' (align first indent in the documentation).",
            )
        return doc_mode == "skip"

    def visit_If(self, node):  # noqa
        # ignore inline ifs and their else/else if branches
        if self.is_inline:
            return node
        self.create_auto_widths_for_context(node)
        self.is_inline = InlineIfHeader and isinstance(node.header, InlineIfHeader)
        if self.is_inline:
            self.is_inline = False
            return node
        if not isinstance(node.header, (ElseHeader, ElseIfHeader)):
            self.indent += 1
        self.generic_visit(node)
        if not isinstance(node.header, (ElseHeader, ElseIfHeader)):
            self.indent -= 1
        self.remove_auto_widths_for_context()
        return node

    def visit_Try(self, node):  # noqa
        self.create_auto_widths_for_context(node)
        # do not increase header for Except, Else, Finally - it was done in Try already
        if isinstance(node.header, TryHeader):
            self.indent += 1
        self.generic_visit(node)
        if isinstance(node.header, TryHeader):
            self.indent -= 1
        self.remove_auto_widths_for_context()
        return node

    def visit_For(self, node):  # noqa
        self.create_auto_widths_for_context(node)
        self.indent += 1
        self.generic_visit(node)
        self.indent -= 1
        self.remove_auto_widths_for_context()
        return node

    visit_While = visit_For

    def get_width(self, col, override_default_zero=False):
        # If auto mode is enabled, use auto widths for current context (last defined widths)
        if self.auto_widths:
            widths = self.auto_widths[-1]
        else:
            widths = self.widths
        if not widths:
            return self.DEFAULT_WIDTH
        if col in widths:
            return widths[col]
        width = widths[len(widths) - 1]  # if there is no such column, use last column width
        if override_default_zero and width == 0:  # edge case where 0 is last of widths and we're overflowing
            return self.formatting_config.space_count
        return width

    def visit_SettingSection(self, node):  # noqa
        return node

    @skip_if_disabled
    def visit_Documentation(self, node):  # noqa
        if self.skip.documentation:
            return node
        # For every line:
        # {indent}...{aligned separator}{leave rest alone}
        width = self.get_width(0)
        for line in node.lines:
            first_sep = True
            prev_token = None
            for token in line:
                if token.type == Token.SEPARATOR:
                    if first_sep:
                        token.value = self.formatting_config.indent
                        first_sep = False
                        continue
                    if width == 0:
                        separator_len = round_to_four(len(prev_token.value) + self.formatting_config.space_count) - len(
                            prev_token.value
                        )
                    else:
                        separator_len = max(width - len(prev_token.value), self.formatting_config.space_count)
                    token.value = " " * separator_len
                    break
                elif token.type != Token.ARGUMENT:  # ...   # comment edge case
                    prev_token = token
        return node

    def create_auto_widths_for_context(self, node):
        if self.fixed_alignment:
            return
        counter = ColumnWidthCounter(
            self.disablers,
            self.skip.documentation,
            self.handle_too_long,
            self.widths,
            self.DEFAULT_WIDTH,
            self.formatting_config.space_count,
        )
        counter.visit(node)
        counter.calculate_column_widths()
        self.auto_widths.append(counter.widths)

    def remove_auto_widths_for_context(self):
        if not self.fixed_alignment:
            self.auto_widths.pop()

    def visit_ForHeader(self, node):  # noqa
        # Fix indent for FOR, IF, WHILE, TRY block headers & ends
        indent = Token(Token.SEPARATOR, (self.indent - 1) * self.formatting_config.indent)
        node.tokens = [indent] + list(node.tokens[1:])
        return node

    visit_End = visit_ForHeader  # TODO add other headers

    @skip_if_disabled
    def visit_KeywordCall(self, node):  # noqa
        if node.errors:
            return node
        if self.skip.keyword_call(node):
            return node
        return self.align_node(node, check_length=self.split_too_long, possible_assign=True)

    def should_skip_return_values(self, line, possible_assign):
        return possible_assign and self.skip.return_values and any(token.type == Token.ASSIGN for token in line)

    def align_node(self, node, check_length: bool, possible_assign: bool = False):
        indent = Token(Token.SEPARATOR, self.indent * self.formatting_config.indent)
        aligned_lines = []
        for line in node.lines:
            assign, tokens, skip_width = self.split_assign(line, possible_assign)
            aligned_line = self.align_line(tokens, skip_width)
            if aligned_line is None:
                aligned_lines.extend(line)
                continue
            aligned_line = [indent] + assign + aligned_line
            if check_length and self.is_line_too_long(aligned_line):
                split_node = self.split_too_long_node(node)
                return self.align_node(split_node, check_length=False)
            aligned_lines.extend(aligned_line)
        return Statement.from_tokens(aligned_lines)

    def split_assign(self, line: List, possible_assign: bool) -> Tuple[List, List, int]:
        """
        This method returns return values together with their separators in case
        we don't want to align them.

        Returns:
            A tuple, containing:
            - return values,
            - remaining tokens,
            - widths of the return values (used to determine next alignment column)
        """
        if not self.should_skip_return_values(line, possible_assign):
            return [], get_data_tokens(line), 0
        assign = []
        assign_found = False
        for index, token in enumerate(line[1:], start=1):
            if token.type == Token.ASSIGN:
                assign_found = True
            elif assign_found and token.type not in (Token.SEPARATOR, Token.ASSIGN):
                skip_width = sum(len(token.value) for token in assign[:-1])
                return (
                    assign[:-1],
                    get_data_tokens(line[index:]),
                    skip_width,
                )
            assign.append(token)

    @skip_if_disabled
    def visit_Tags(self, node):  # noqa
        if node.errors or self.skip.setting("Tags"):
            return node
        return self.align_node(node, check_length=False)

    @skip_if_disabled
    def visit_Return(self, node):  # noqa
        if node.errors or self.skip.setting("Return_Statement"):
            return node
        return self.align_node(node, check_length=False)

    visit_ReturnStatement = visit_Return

    @skip_if_disabled
    def visit_Template(self, node):  # noqa
        if node.errors or self.skip.setting("Template"):
            return node
        return self.align_node(node, check_length=False)

    @skip_if_disabled
    def visit_Timeout(self, node):  # noqa
        if node.errors or self.skip.setting("Timeout"):
            return node
        return self.align_node(node, check_length=False)

    @skip_if_disabled
    def visit_Teardown(self, node):  # noqa
        if node.errors or self.skip.setting("Teardown"):
            return node
        return self.align_node(node, check_length=False)

    @skip_if_disabled
    def visit_Setup(self, node):  # noqa
        if node.errors or self.skip.setting("Setup"):
            return node
        return self.align_node(node, check_length=False)

    @skip_if_disabled
    def visit_Arguments(self, node):  # noqa
        if node.errors or self.skip.setting("Arguments"):
            return node
        return self.align_node(node, check_length=False)

    def align_line(self, line: List, skip_width: int):
        """
        Align single line of the node.

        New, aligned line consist of the indent and tokens separated by separator. The separator width is calculated
        depending on the configured particular column width and the alignment mode.

        The alignment mode could be auto or fixed - it's calculated before running this method and self.widths already
        returns width of the column according to the selected mode.

        First, empty multiline is return as it can't be aligned.
        Then comments are removed from the line - they will be added at the end of the alignment.
        Afterwards we start to align the tokens.
        If the configured width is set to 0 it means we don't need to calculate width of the separator - we should use
        length of the token with minimal separator rounded to the closest multiply of 4.
        Otherwise, we are calculating if line will fit in width of the column. In case it will not fit there are
        several options (configurable via ``handle_too_long):
            - ignore_line - the whole line will be separated using fixed, minimal width of the separator
            - ignore_rest - all tokens before too long token will be aligned to the columns, but remaining tokens will
              be aligned using fixed, minimal width of the separator
            - compact_overflow - the too long token will go over the width of the column, but it will try to fit it
              together with next token (in second column)
            - overflow - we are looking for next column width that will fit our token

        At the end comments are appended at the end of the line and line is returned.
        """
        if is_blank_multiline(line):  # ...\n edge case
            line[-1].value = line[-1].value.lstrip(" \t")  # normalize eol from '  \n' to '\n'
            return line
        tokens, comments = separate_comments(line)
        if len(tokens) < 2:  # only happens with weird encoding, better to skip
            return None
        # skip_tokens, tokens = self.skip_return_values_if_needed(tokens)
        aligned = self.align_tokens(tokens[:-2], skip_width)
        last_token = strip_extra_whitespace(tokens[-2])
        aligned.extend([last_token, *join_comments(comments), tokens[-1]])
        return aligned

    def too_many_misaligned_cols(self, misaligned_cols: int, prev_overflow_len: int, tokens: List, index: int):
        return misaligned_cols >= self.compact_overflow_limit and prev_overflow_len and index < len(tokens) - 1

    def find_starting_column(self, skip_width: int):
        """
        If we're skipping values at the beginning of the line,
        we need to find next column for remaining tokens.
        """
        column = 0
        while skip_width > 0:
            width = self.get_width(column, override_default_zero=True)
            skip_width -= width
            column += 1
        return column, abs(skip_width)

    def get_start_column_and_aligned(self, skip_width: int, min_separator: int):
        """
        In case we are skipping return tokens alignment, calculate starting column and create leading separator.
        """
        if skip_width == 0:
            return 0, 0, []
        column, skip_width = self.find_starting_column(skip_width)
        if skip_width < min_separator:
            prev_overflow_len = min_separator - skip_width
            skip_width = min_separator
        else:
            prev_overflow_len = 0
        return column, prev_overflow_len, [get_separator(skip_width)]

    def align_tokens(self, tokens: List, skip_width: int):
        last_assign, misaligned_cols = 0, 0
        min_separator = self.formatting_config.space_count
        column, prev_overflow_len, aligned = self.get_start_column_and_aligned(skip_width, min_separator)
        for index, token in enumerate(tokens):
            aligned.append(token)
            width = self.get_width(column)
            if width == 0:
                separator_len = round_to_four(len(token.value) + min_separator) - len(token.value)
            else:
                separator_len = width - len(token.value) - prev_overflow_len
                if separator_len >= min_separator:
                    prev_overflow_len = 0
                    misaligned_cols = 0
                else:
                    if self.handle_too_long == "ignore_line":
                        return align_fixed(tokens, min_separator)
                    if self.handle_too_long == "ignore_rest":
                        return aligned + align_fixed(tokens[index + 1 :], min_separator, start_sep=True)
                    if self.handle_too_long == "compact_overflow":
                        required_width = len(token.value) + min_separator + prev_overflow_len
                        separator_len = min_separator
                        prev_overflow_len = required_width - width
                        misaligned_cols += 1
                        while prev_overflow_len > width:
                            column += 1
                            width = self.get_width(column, override_default_zero=True)
                            prev_overflow_len -= width
                            misaligned_cols += 1
                        if self.too_many_misaligned_cols(misaligned_cols, prev_overflow_len, tokens, index):
                            # check if next col fits next token with prev_overflow, if not, jump to the next column
                            next_token = tokens[index + 1]
                            next_width = self.get_width(column + 1, override_default_zero=True)
                            required_width = next_width - prev_overflow_len - len(next_token.value)
                            if required_width < min_separator:
                                column += 1
                                separator_len = next_width - prev_overflow_len + min_separator
                                prev_overflow_len = 0
                    else:  # "overflow"
                        while round_to_four(len(token.value) + min_separator) > width:
                            column += 1
                            width += self.get_width(column, override_default_zero=True)
                        separator_len = width - len(token.value)
            separator_len = max(
                min_separator, separator_len
            )  # extra precautious: separator should never be less than min_sep
            aligned.append(get_separator(separator_len))
            column += 1 + (separator_len == width)
        return aligned

    def is_line_too_long(self, line):
        if "SplitTooLongLine" not in self.transformers:
            return False
        if not self.transformers["SplitTooLongLine"].split_on_every_arg:  # TODO not support for overflow yet
            return False
        line_length = get_line_length(line)
        return line_length > self.transformers["SplitTooLongLine"].line_length

    def is_split_too_long_enabled(self):
        return "SplitTooLongLine" in self.transformers

    def split_too_long_node(self, node):
        return self.transformers["SplitTooLongLine"].split_keyword_call(node)


def strip_extra_whitespace(token):
    if not token.value:
        return token
    token.value = token.value.strip()
    return token


def separate_comments(tokens):
    non_comments, comments = [], []
    for token in tokens:
        if token.type == Token.COMMENT:
            comments.append(token)
        else:
            non_comments.append(token)
    return non_comments, comments


def join_comments(comments):
    tokens = []
    separator = get_separator(2)
    for token in comments:
        tokens.append(separator)
        tokens.append(token)
    return tokens


def align_fixed(tokens, sep_len, start_sep=False):
    """Align tokens with fixed spacing."""
    sep_token = get_separator(sep_len)
    aligned = [sep_token] if start_sep else []
    for token in tokens:
        aligned.append(token)
        aligned.append(sep_token)
    return aligned


def get_line_length(tokens):
    return sum(len(token.value) for token in tokens)


def get_data_tokens(tokens):
    return [token for token in tokens if token.type not in WHITESPACE_TOKENS]


def get_separator(sep_len: int) -> Token:
    return Token(Token.SEPARATOR, sep_len * " ")


class ColumnWidthCounter(ModelVisitor):
    NON_DATA_TOKENS = frozenset((Token.SEPARATOR, Token.COMMENT, Token.EOL, Token.EOS))

    def __init__(self, disablers, skip_documentation, handle_too_long, max_widths, default_width, min_separator):
        self.skip_documentation = skip_documentation
        self.handle_too_long = handle_too_long
        self.max_widths = max_widths
        self.default_width = default_width
        self.min_separator = min_separator
        self.raw_widths = defaultdict(list)
        self.widths = dict()
        self.disablers = disablers

    def get_width(self, col):
        if not self.max_widths:
            return self.default_width
        if col in self.max_widths:
            return self.max_widths[col]
        return self.max_widths[len(self.max_widths) - 1]  # if there is no such column, use last column width

    def calculate_column_widths(self):
        if self.max_widths:
            self.widths.update(self.max_widths)
        for column, widths in self.raw_widths.items():
            max_width = self.get_width(column)
            if max_width == 0:
                self.widths[column] = max(widths)
            else:
                filter_widths = [width for width in widths if width <= max_width]
                self.widths[column] = max(filter_widths, default=max_width)

    @skip_if_disabled
    def visit_KeywordCall(self, node):  # noqa
        if node.errors:
            return node
        for line in node.lines:
            # if assign disabled and assign in line: continue
            data_tokens = [token for token in line if token.type not in self.NON_DATA_TOKENS]
            raw_lens = {}
            for column, token in enumerate(data_tokens):
                max_width = self.get_width(column)
                token_len = round_to_four(len(token.value) + self.min_separator)
                if max_width == 0 or token_len <= max_width:
                    raw_lens[column] = token_len
                elif self.handle_too_long == "ignore_line":
                    raw_lens = {}
                    break
                else:  # ignore_rest, overflow and compact_overflow
                    break
            for col, token in raw_lens.items():
                self.raw_widths[col].append(token)

    visit_Arguments = (
        visit_Setup
    ) = visit_Teardown = visit_Timeout = visit_Template = visit_Return = visit_Tags = visit_KeywordCall  # TODO skip

    @skip_if_disabled
    def visit_Documentation(self, node):  # noqa
        if self.skip_documentation:
            return
        doc_header_len = round_to_four(len(node.data_tokens[0].value) + self.min_separator)
        self.raw_widths[0].append(doc_header_len)



import datetime
import json


def _decode_oid(decoder, oid):
    return decoder.memo[oid]


def _decode_float(decoder, msg):
    return float(msg)


def _decode_int(decoder, msg):
    return int(msg)


def _decode_str(decoder, msg):
    return msg


red = "\u001b[31m"
green = "\u001b[32m"
yellow = "\u001b[33m"
blue = "\u001b[34m"
magenta = "\u001b[35m"
cyan = "\u001b[36m"
white = "\u001b[37m"
reset = "\u001b[0m"


def _color_status(status):
    if status == "PASS":
        return f"{green}{status}{reset}"
    elif status == "NOT RUN":
        return f"{cyan}{status}{reset}"
    else:
        return f"{red}{status}{reset}"


def _color_name(status):
    return f"{yellow}{status}{reset}"


def _decode(message_definition, level_diff=0):
    names = []
    name_to_decode = {}
    for s in message_definition.split(","):
        s = s.strip()
        i = s.find(":")
        decode = "oid"
        if i != -1:
            s, decode = s.split(":", 1)
        names.append(s)
        if decode == "oid":
            name_to_decode[s] = _decode_oid

        elif decode == "int":
            name_to_decode[s] = _decode_int

        elif decode == "float":
            name_to_decode[s] = _decode_float

        elif decode == "str":
            name_to_decode[s] = _decode_str

        else:
            raise RuntimeError(f"Unexpected: {decode}")

    def dec_impl(decoder, message):
        decoder.level += level_diff
        splitted = message.split("|", len(names) - 1)
        ret = {}
        for i, s in enumerate(splitted):
            name = names[i]
            try:
                ret[name] = name_to_decode[name](decoder, s)
            except:
                ret[name] = None
        return ret

    return dec_impl


def decode_time(decoder, time):
    decoder.initial_time = datetime.datetime.fromisoformat(time)
    return {"initial_time": time}


def decode_memo(decoder, message):
    memo_id, memo_value = message.split(":", 1)
    memo_value = json.loads(memo_value)
    decoder.memo[memo_id] = memo_value
    return None


_MESSAGE_TYPE_INFO = {
    "V": lambda _decoder, message: {"version": message},
    "I": lambda _decoder, message: {"info": json.loads(message)},
    "ID": _decode("part:int, id:str"),
    "T": decode_time,
    "M": decode_memo,
    "L": _decode("level:str, message:oid, time_delta_in_seconds:float"),
    "LH": _decode("level:str, message:oid, time_delta_in_seconds:float"),
    "SS": _decode(
        "name:oid, suite_id:oid, suite_source:oid, time_delta_in_seconds:float",
        level_diff=+1,
    ),
    "ES": _decode("status:oid, time_delta_in_seconds:float", level_diff=-1),
    "ST": _decode(
        "name:oid, suite_id:oid, lineno:int, time_delta_in_seconds:float", level_diff=+1
    ),
    "ET": _decode(
        "status:oid, message:oid, time_delta_in_seconds:float", level_diff=-1
    ),
    "SK": _decode(
        "name:oid, libname:oid, keyword_type:oid, doc:oid, source:oid, lineno:int, time_delta_in_seconds:float",
        level_diff=+1,
    ),
    "EK": _decode("status:oid, time_delta_in_seconds:float", level_diff=-1),
    "KA": _decode("argument:oid"),
    "AS": _decode("assign:oid"),
    "TG": _decode("tag:oid"),
    "S": _decode("start_time_delta:float"),
}

_MESSAGE_TYPE_INFO["RS"] = _MESSAGE_TYPE_INFO["SS"]
_MESSAGE_TYPE_INFO["RT"] = _MESSAGE_TYPE_INFO["ST"]
_MESSAGE_TYPE_INFO["RK"] = _MESSAGE_TYPE_INFO["SK"]


class Decoder:
    def __init__(self):
        self.memo = {}
        self.initial_time = None
        self.level = 0

    @property
    def ident(self):
        return "    " * self.level

    def decode_message_type(self, message_type, message):
        handler = _MESSAGE_TYPE_INFO[message_type]
        ret = {"message_type": message_type}
        try:
            r = handler(self, message)
            if not r:
                if message_type == "M":
                    return None
                raise RuntimeError(
                    f"No return when decoding: {message_type} - {message}"
                )
                if not isinstance(r, dict):
                    ret[
                        "error"
                    ] = f"Expected dict return when decoding: {message_type} - {message}. Found: {ret}"

            ret.update(r)
        except Exception as e:
            ret["error"] = f"Error decoding: {message_type}: {e}"
        return ret


def iter_decoded_log_format(stream):
    decoder = Decoder()
    for line in stream.readlines():
        line = line.strip()
        if line:
            message_type, message = line.split(" ", 1)
            decoded = decoder.decode_message_type(message_type, message)
            if decoded:
                yield decoded



from robot.api.parsing import Comment, EmptyLine

try:
    from robot.api.parsing import ReturnStatement
except ImportError:
    ReturnStatement = None

from robotidy.disablers import skip_if_disabled, skip_section_if_disabled
from robotidy.transformers import Transformer
from robotidy.utils import (
    after_last_dot,
    create_statement_from_tokens,
    normalize_name,
    wrap_in_if_and_replace_statement,
)


class ReplaceReturns(Transformer):
    """
    Replace return statements (such as [Return] setting or Return From Keyword keyword) with RETURN statement.

    Following code:

    ```robotframework
    *** Keywords ***
    Keyword
        Return From Keyword If    $condition    2
        Sub Keyword
        [Return]    1

    Keyword 2
        Return From Keyword    ${arg}
    ```

    will be transformed to:

    ```robotframework
    *** Keywords ***
    Keyword
        IF    $condition
            RETURN    2
        END
        Sub Keyword
        RETURN    1

    Keyword 2
        RETURN    ${arg}
    ```
    """

    MIN_VERSION = 5

    def __init__(self):
        super().__init__()
        self.return_statement = None

    @skip_section_if_disabled
    def visit_Section(self, node):  # noqa
        return self.generic_visit(node)

    def visit_Keyword(self, node):  # noqa
        self.return_statement = None
        node = self.generic_visit(node)
        if self.return_statement:
            skip_lines = []
            indent = self.return_statement.tokens[0]
            while node.body and isinstance(node.body[-1], (EmptyLine, Comment)):
                skip_lines.append(node.body.pop())
            return_stmt = create_statement_from_tokens(
                statement=ReturnStatement, tokens=self.return_statement.tokens[2:], indent=indent
            )
            node.body.append(return_stmt)
            node.body.extend(skip_lines)
        return node

    @skip_if_disabled
    def visit_KeywordCall(self, node):  # noqa
        if not node.keyword or node.errors:
            return node
        normalized_name = after_last_dot(normalize_name(node.keyword))
        if normalized_name == "returnfromkeyword":
            return create_statement_from_tokens(
                statement=ReturnStatement, tokens=node.tokens[2:], indent=node.tokens[0]
            )
        elif normalized_name == "returnfromkeywordif":
            return wrap_in_if_and_replace_statement(node, ReturnStatement, self.formatting_config.separator)
        return node

    @skip_if_disabled
    def visit_Return(self, node):  # noqa
        self.return_statement = node

    @skip_if_disabled
    def visit_Error(self, node):  # noqa
        """Remove duplicate [Return]"""
        for error in node.errors:
            if "Setting 'Return' is allowed only once" in error:
                return None
        return node



from typing import Tuple, Set, Union, Iterator, Optional, Any
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.protocols import ILocalizationInfo

log = get_logger(__name__)


def _get_lang_from_code(language_code, __code_to_language_cache={}):
    try:
        return __code_to_language_cache[language_code]
    except KeyError:
        from robot.api import Language

        try:
            __code_to_language_cache[language_code] = Language.from_name(language_code)
        except:
            log.exception(f"Unable to load language: {language_code}")
            # Cache as None
            __code_to_language_cache[language_code] = None

        return __code_to_language_cache[language_code]


class LocalizationInfo:
    def __init__(self, language_codes: Union[Tuple[str, ...], str]):
        if isinstance(language_codes, str):
            language_codes = (language_codes,)
        self._language_codes: Tuple[str, ...] = language_codes

        self._last_bdd_prefixes_cache_key: Optional[Set[str]] = None
        self._bdd_prefixes: Optional[Set[str]] = None

    def __str__(self):
        return f"LocalizationInfo({self._language_codes})"

    @property
    def language_codes(self) -> Tuple[str, ...]:
        return self._language_codes  # Note: could actually be language code or name.

    def iter_languages_on_write(
        self,
    ) -> Iterator[Any]:  # Actually Iterator[robot.api.Language]
        from robotframework_ls.impl.robot_version import robot_version_supports_language

        if robot_version_supports_language():
            languages = set()

            found = False

            if self._language_codes:
                # i.e.: prefer language specified locally.
                for lang_code in set(self._language_codes):
                    lang = _get_lang_from_code(lang_code)
                    if lang is None:
                        log.info("Could not find language: %s", lang_code)
                    else:
                        if lang not in languages:
                            languages.add(lang)
                            found = True
                            yield lang
                if found:
                    return

            # Not found locally: fall back to global.
            global_localization_info = get_global_localization_info()

            if global_localization_info.language_codes:
                for lang_code in set(global_localization_info.language_codes):
                    lang = _get_lang_from_code(lang_code)
                    if lang is None:
                        log.info(
                            "Could not find language: %s (set in global)", lang_code
                        )
                    else:
                        if lang not in languages:
                            languages.add(lang)
                            found = True
                            yield lang

                if found:
                    return

            # No language defined locally nor globally: just use english.
            lang = _get_lang_from_code("en")
            if lang is None:
                log.critical("en lang returning None (this should not be possible).")
            else:
                yield lang

    def iter_bdd_prefixes_on_read(self) -> Iterator[str]:
        """
        Note that we specify the reason for iterating because for instance, when
        writing code we could want just the completions for the specified
        language in the file and while reading (i.e.: analyzing) we'd want it
        for all languages.
        """
        from robotframework_ls.impl.robot_version import robot_version_supports_language

        global_localization_info = get_global_localization_info()
        global_language_codes = set(global_localization_info.language_codes)
        if self._last_bdd_prefixes_cache_key != global_language_codes:
            self._bdd_prefixes = None  # Clear cache

        if self._bdd_prefixes is None:
            from robotframework_ls.impl.robot_constants import BDD_PREFIXES

            bdd_prefixes: Set[str] = set()
            bdd_prefixes.update(BDD_PREFIXES)

            if robot_version_supports_language():
                use_language_codes = set(global_language_codes)
                use_language_codes.update(self._language_codes)

                for lang_code in use_language_codes:
                    if lang_code.lower() in ("en", "english"):
                        # Just for English use our constants (which are already loaded).
                        continue

                    else:
                        lang = _get_lang_from_code(lang_code)
                        if lang is not None:
                            for prefix in iter(lang.bdd_prefixes):
                                # Note: iterating over the same language multiple
                                # times would be ok here.
                                bdd_prefixes.add(prefix.lower())

            self._last_bdd_prefixes_cache_key = global_language_codes
            self._bdd_prefixes = bdd_prefixes

        yield from iter(self._bdd_prefixes)

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: ILocalizationInfo = check_implements(self)


class _LocalizationInfoHolder:
    localization_info = LocalizationInfo("en")  # Default is just English.


def set_global_localization_info(localization_info: LocalizationInfo):
    log.debug(
        "Setting global localization language: %s", localization_info.language_codes
    )
    _LocalizationInfoHolder.localization_info = localization_info


def get_global_localization_info() -> LocalizationInfo:
    return _LocalizationInfoHolder.localization_info


def set_global_from_config(config):
    from robotframework_ls.impl.robot_generated_lsp_constants import (
        OPTION_ROBOT_LANGUAGE,
    )

    try:
        language_codes = config.get_setting(OPTION_ROBOT_LANGUAGE, list, [])
        if not language_codes:
            language_codes = "en"
        set_global_localization_info(LocalizationInfo(language_codes))

    except:
        log.exception(f"Error setting localization info from configuration.")
        # Fallback to English.
        set_global_localization_info(LocalizationInfo("en"))



# Original work Copyright 2017 Palantir Technologies, Inc. (MIT)
# See ThirdPartyNotices.txt in the project root for license information.
# All modifications Copyright (c) Robocorp Technologies Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: // www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import os
import sys
import threading

from contextlib import contextmanager
from robocorp_ls_core.robotframework_log import get_logger
from robocorp_ls_core.options import DEFAULT_TIMEOUT
from typing import TypeVar, Any, Callable, Tuple
from robocorp_ls_core.jsonrpc.exceptions import JsonRpcRequestCancelled
from functools import lru_cache


PARENT_PROCESS_WATCH_INTERVAL = 3  # 3 s


def as_str(s) -> str:
    if isinstance(s, bytes):
        return s.decode("utf-8", "replace")
    return str(s)


log = get_logger(__name__)


def list_to_string(value):
    return ",".join(value) if isinstance(value, list) else value


if sys.platform == "win32":
    import ctypes

    kernel32 = ctypes.windll.kernel32
    PROCESS_SYNCHRONIZE = 0x00100000
    DWORD = ctypes.c_uint32
    BOOL = ctypes.c_int
    LPVOID = ctypes.c_void_p
    HANDLE = LPVOID

    OpenProcess = kernel32.OpenProcess
    OpenProcess.argtypes = [DWORD, BOOL, DWORD]
    OpenProcess.restype = HANDLE

    WaitForSingleObject = kernel32.WaitForSingleObject
    WaitForSingleObject.argtypes = [HANDLE, DWORD]
    WaitForSingleObject.restype = DWORD

    WAIT_TIMEOUT = 0x00000102
    WAIT_ABANDONED = 0x00000080
    WAIT_OBJECT_0 = 0
    WAIT_FAILED = 0xFFFFFFFF

    def is_process_alive(pid):
        """Check whether the process with the given pid is still alive.

        Running `os.kill()` on Windows always exits the process, so it can't be used to check for an alive process.
        see: https://docs.python.org/3/library/os.html?highlight=os%20kill#os.kill

        Hence ctypes is used to check for the process directly via windows API avoiding any other 3rd-party dependency.

        Args:
            pid (int): process ID

        Returns:
            bool: False if the process is not alive or don't have permission to check, True otherwise.
        """
        process = OpenProcess(PROCESS_SYNCHRONIZE, 0, pid)
        if process != 0:
            try:
                wait_result = WaitForSingleObject(process, 0)
                if wait_result == WAIT_TIMEOUT:
                    return True
            finally:
                kernel32.CloseHandle(process)
        return False


else:
    import errno

    def _is_process_alive(pid):
        """Check whether the process with the given pid is still alive.

        Args:
            pid (int): process ID

        Returns:
            bool: False if the process is not alive or don't have permission to check, True otherwise.
        """
        if pid < 0:
            return False
        try:
            os.kill(pid, 0)
        except OSError as e:
            if e.errno == errno.ESRCH:
                return False  # No such process.
            elif e.errno == errno.EPERM:
                return True  # permission denied.
            else:
                log.info("Unexpected errno: %s", e.errno)
                return False
        else:
            return True

    def is_process_alive(pid):
        from robocorp_ls_core.subprocess_wrapper import subprocess

        if _is_process_alive(pid):
            # Check if zombie...
            try:
                cmd = ["ps", "-p", str(pid), "-o", "stat"]
                try:
                    process = subprocess.Popen(
                        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
                    )
                except:
                    log.exception("Error calling: %s.", " ".join(cmd))
                else:
                    stdout, _ = process.communicate()
                    stdout = stdout.decode("utf-8", "replace")
                    lines = [line.strip() for line in stdout.splitlines()]
                    if len(lines) > 1:
                        if lines[1].startswith("Z"):
                            return False  # It's a zombie
            except:
                log.exception("Error checking if process is alive.")

            return True
        return False


def _popen(cmdline, **kwargs):
    import subprocess

    try:
        return subprocess.Popen(cmdline, **kwargs)
    except:
        log.exception("Error running: %s", (" ".join(cmdline)))
        return None


def _call(cmdline, **kwargs):
    import subprocess

    try:
        subprocess.check_call(cmdline, **kwargs)
    except:
        log.exception("Error running: %s", (" ".join(cmdline)))
        return None


def _kill_process_and_subprocess_linux(pid):
    import subprocess

    initial_pid = pid

    def list_children_and_stop_forking(ppid):
        children_pids = []
        _call(
            ["kill", "-STOP", str(ppid)], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        list_popen = _popen(
            ["pgrep", "-P", str(ppid)], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        if list_popen is not None:
            stdout, _ = list_popen.communicate()
            for line in stdout.splitlines():
                line = line.decode("ascii").strip()
                if line:
                    pid = str(line)
                    children_pids.append(pid)
                    # Recursively get children.
                    children_pids.extend(list_children_and_stop_forking(pid))
        return children_pids

    previously_found = set()

    for _ in range(50):  # Try this at most 50 times before giving up.
        children_pids = list_children_and_stop_forking(initial_pid)
        found_new = False

        for pid in children_pids:
            if pid not in previously_found:
                found_new = True
                previously_found.add(pid)
                _call(
                    ["kill", "-KILL", str(pid)],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )

        if not found_new:
            break

    # Now, finish the initial one.
    _call(
        ["kill", "-KILL", str(initial_pid)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )


def kill_process_and_subprocesses(pid):
    log.debug("Killing process and subprocesses of: %s", pid)
    from subprocess import CalledProcessError

    if sys.platform == "win32":
        from robocorp_ls_core.subprocess_wrapper import subprocess

        args = ["taskkill", "/F", "/PID", str(pid), "/T"]
        retcode = subprocess.call(
            args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE
        )
        if retcode not in (0, 128, 255):
            raise CalledProcessError(retcode, args)
    else:
        _kill_process_and_subprocess_linux(pid)


_track_pids_to_exit = set()
_watching_thread_global = None


def exit_when_pid_exists(pid):
    _track_pids_to_exit.add(pid)
    global _watching_thread_global
    if _watching_thread_global is None:
        import time

        def watch_parent_process():
            # exit when any of the ids we're tracking exit.
            while True:
                for pid in _track_pids_to_exit:
                    if not is_process_alive(pid):
                        # Note: just exit since the parent process already
                        # exited.
                        log.info(
                            f"Force-quit process: %s because parent: %s exited",
                            os.getpid(),
                            pid,
                        )
                        os._exit(0)

                time.sleep(PARENT_PROCESS_WATCH_INTERVAL)

        _watching_thread_global = threading.Thread(target=watch_parent_process, args=())
        _watching_thread_global.daemon = True
        _watching_thread_global.start()


F = TypeVar("F", bound=Callable[..., Any])


def overrides(method: Any) -> Callable[[F], F]:
    """
    Meant to be used as

    class B:
        @overrides(A.m1)
        def m1(self):
            pass
    """

    @functools.wraps(method)
    def wrapper(func):
        if func.__name__ != method.__name__:
            msg = f"Wrong @override: {func.__name__!r} expected, but overwriting {method.__name__!r}."
            raise AssertionError(msg)

        return func

    return wrapper


def implements(method: Any) -> Callable[[F], F]:
    @functools.wraps(method)
    def wrapper(func):
        if func.__name__ != method.__name__:
            msg = f"Wrong @implements: {func.__name__!r} expected, but implementing {method.__name__!r}."
            raise AssertionError(msg)

        return func

    return wrapper


def log_and_silence_errors(logger, return_on_error=None):
    def inner(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except JsonRpcRequestCancelled:
                logger.info("Cancelled handling: %s", func)
                raise  # Don't silence cancelled exceptions
            except:
                logger.exception("Error calling: %s", func)
                return return_on_error

        return new_func

    return inner


def log_but_dont_silence_errors(logger):
    def inner(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except JsonRpcRequestCancelled:
                logger.info("Cancelled handling: %s", func)
                raise  # Don't silence cancelled exceptions
            except:
                logger.exception("Error calling: %s", func)
                raise

        return new_func

    return inner


@contextmanager
def after(obj, method_name, callback):
    original_method = getattr(obj, method_name)

    @functools.wraps(original_method)
    def new_method(*args, **kwargs):
        ret = original_method(*args, **kwargs)
        callback(*args, **kwargs)
        return ret

    setattr(obj, method_name, new_method)
    try:
        yield
    finally:
        setattr(obj, method_name, original_method)


@contextmanager
def before(obj, method_name, callback):
    original_method = getattr(obj, method_name)

    @functools.wraps(original_method)
    def new_method(*args, **kwargs):
        callback(*args, **kwargs)
        ret = original_method(*args, **kwargs)
        return ret

    setattr(obj, method_name, new_method)
    try:
        yield
    finally:
        setattr(obj, method_name, original_method)


def check_min_version(version: str, min_version: Tuple[int, int]) -> bool:
    """
    :param version:
        This is the version of robotframework.

    :param min_version:
        This is the minimum version to match.

    :return bool:
        True if version >= min_versiond and False otherwise.
    """
    try:
        v = tuple(int(x) for x in version.split("."))
    except:
        return False

    return v >= min_version


def wait_for_condition(condition, msg=None, timeout=DEFAULT_TIMEOUT, sleep=1 / 20.0):
    """
    Note: wait_for_expected_func_return is usually a better API to use as
    the error message is automatically built.
    """
    import time

    curtime = time.time()

    while True:
        if condition():
            break
        if timeout is not None and (time.time() - curtime > timeout):
            error_msg = f"Condition not reached in {timeout} seconds"
            if msg is not None:
                error_msg += "\n"
                if callable(msg):
                    error_msg += msg()
                else:
                    error_msg += str(msg)

            raise TimeoutError(error_msg)
        time.sleep(sleep)


def wait_for_non_error_condition(
    generate_error_or_none, timeout=DEFAULT_TIMEOUT, sleep=1 / 20.0
):
    import time

    curtime = time.time()

    while True:
        error_msg = generate_error_or_none()
        if error_msg is None:
            break

        if timeout is not None and (time.time() - curtime > timeout):
            raise TimeoutError(
                f"Condition not reached in {timeout} seconds\n{error_msg}"
            )
        time.sleep(sleep)


def wait_for_expected_func_return(
    func, expected_return, timeout=DEFAULT_TIMEOUT, sleep=1 / 20.0
):
    def check():
        found = func()
        if found != expected_return:
            return "Expected: %s. Found: %s" % (expected_return, found)

        return None

    wait_for_non_error_condition(check, timeout, sleep)


def isinstance_name(obj, classname, memo={}):
    """
    Checks if a given object is instance of a class with the given name.
    """
    if classname.__class__ in (list, tuple):
        for c in classname:
            if isinstance_name(obj, c):
                return True
        return False

    cls = obj.__class__
    key = (cls, classname)
    try:
        return memo[key]
    except KeyError:
        if cls.__name__ == classname:
            memo[key] = True
        else:
            for check in obj.__class__.__mro__:
                if check.__name__ == classname:
                    memo[key] = True
                    break
            else:
                memo[key] = False

        return memo[key]


def build_subprocess_kwargs(cwd, env, **kwargs) -> dict:
    from robocorp_ls_core.subprocess_wrapper import subprocess

    startupinfo = None
    if sys.platform == "win32":
        # We don't want to show the shell on windows!
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        startupinfo.wShowWindow = subprocess.SW_HIDE
        startupinfo = startupinfo

    if cwd:
        kwargs["cwd"] = cwd
    if env:
        kwargs["env"] = env
    kwargs["startupinfo"] = startupinfo
    return kwargs


def make_unique(lst):
    seen = set()
    return [x for x in lst if x not in seen and not seen.add(x)]


@lru_cache(maxsize=3000)
def normalize_filename(filename):
    return os.path.abspath(os.path.normpath(os.path.normcase(filename)))


class _RestoreCtxManager(object):
    def __init__(self, original_import):
        self._original_import = original_import

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        import builtins

        builtins.__import__ = self._original_import


def notify_about_import(import_name):
    """
    :param str import_name:
        The name of the import we don't want in this process.

    Use case: `robot` should not be imported in the Robot Framework Language
    Server process. It should only be imported in the subprocess which is
    spawned specifically for that robot framework version (we should only parse
    the AST at those subprocesses -- if the import is done at the main process
    something needs to be re-engineered to forward the request to a subprocess).

    If used as a context manager restores the previous __import__.
    """
    import builtins

    original_import = builtins.__import__

    import_name_with_dot = import_name + "."

    def new_import(name, *args, **kwargs):
        if name == import_name or name.startswith(import_name_with_dot):
            from io import StringIO
            import traceback

            stream = StringIO()
            stream.write(f"'{name}' should not be imported in this process.\nStack:\n")

            traceback.print_stack(file=stream)

            log.critical(stream.getvalue())

        return original_import(name, *args, **kwargs)

    builtins.__import__ = new_import
    return _RestoreCtxManager(original_import)



from robot.api.parsing import Comment, EmptyLine, Token

from robotidy.disablers import skip_if_disabled, skip_section_if_disabled
from robotidy.exceptions import InvalidParameterValueError, RobotidyConfigError
from robotidy.transformers import Transformer


class InvalidSettingsOrderError(InvalidParameterValueError):
    def __init__(self, transformer, param_name, param_value, valid_values):
        valid_names = ",".join(sorted(valid_values.keys()))
        msg = f"Custom order should be provided in comma separated list with valid setting names: {valid_names}"
        super().__init__(transformer, param_name, param_value, msg)


class DuplicateInSettingsOrderError(InvalidParameterValueError):
    def __init__(self, transformer, param_name, param_value):
        provided_order = ",".join(param.lower() for param in param_value)
        msg = "Custom order cannot contain duplicated setting names."
        super().__init__(transformer, param_name, provided_order, msg)


class SettingInBothOrdersError(RobotidyConfigError):
    def __init__(self, transformer, first_order, second_order, duplicates):
        names = ",".join(setting.lower() for setting in duplicates)
        msg = f"{transformer}: Invalid '{first_order}' and '{second_order}' order values. Following setting names exists in both orders: {names}"
        super().__init__(msg)


class OrderSettings(Transformer):
    """
    Order settings like ``[Arguments]``, ``[Setup]``, ``[Return]`` inside Keywords and Test Cases.

    Keyword settings ``[Documentation]``, ``[Tags]``, ``[Timeout]``, ``[Arguments]`` are put before keyword body and
    settings like ``[Teardown]``, ``[Return]`` are moved to the end of the keyword:

    ```robotframework
    *** Keywords ***
    Keyword
        [Teardown]  Keyword
        [Return]  ${value}
        [Arguments]  ${arg}
        [Documentation]  this is
        ...    doc
        [Tags]  sanity
        Pass
    ```

    To:
    ```robotframework
    *** Keywords ***
    Keyword
        [Documentation]  this is
        ...    doc
        [Tags]  sanity
        [Arguments]  ${arg}
        Pass
        [Teardown]  Keyword
        [Return]  ${value}
    ```

    Test case settings ``[Documentation]``, ``[Tags]``, ``[Template]``, ``[Timeout]``, ``[Setup]`` are put before
    test case body and ``[Teardown]`` is moved to the end of test case.

    Default order can be changed using following parameters:
      - ``keyword_before = documentation,tags,timeout,arguments``
      - ``keyword_after = teardown,return``
      - ``test_before = documentation,tags,template,timeout,setup``
      - ``test_after = teardown``

    Not all settings names need to be passed to given parameter. Missing setting names are not ordered. Example::

        robotidy --configure OrderSettings:keyword_before=:keyword_after=

    It will order only test cases because all setting names for keywords are missing.
    """

    KEYWORD_SETTINGS = {
        "documentation": Token.DOCUMENTATION,
        "tags": Token.TAGS,
        "timeout": Token.TIMEOUT,
        "arguments": Token.ARGUMENTS,
        "return": Token.RETURN,
        "teardown": Token.TEARDOWN,
    }
    TEST_SETTINGS = {
        "documentation": Token.DOCUMENTATION,
        "tags": Token.TAGS,
        "timeout": Token.TIMEOUT,
        "template": Token.TEMPLATE,
        "setup": Token.SETUP,
        "teardown": Token.TEARDOWN,
    }

    def __init__(
        self,
        keyword_before: str = "documentation,tags,timeout,arguments",
        keyword_after: str = "teardown,return",
        test_before: str = "documentation,tags,template,timeout,setup",
        test_after: str = "teardown",
    ):
        super().__init__()
        self.keyword_before = self.get_order(keyword_before, "keyword_before", self.KEYWORD_SETTINGS)
        self.keyword_after = self.get_order(keyword_after, "keyword_after", self.KEYWORD_SETTINGS)
        self.test_before = self.get_order(test_before, "test_before", self.TEST_SETTINGS)
        self.test_after = self.get_order(test_after, "test_after", self.TEST_SETTINGS)
        self.all_keyword_settings = {*self.keyword_before, *self.keyword_after}
        self.all_test_settings = {*self.test_before, *self.test_after}
        self.assert_no_duplicates_in_orders()

    def get_order(self, order, param_name, name_map):
        if not order:
            return []
        parts = order.lower().split(",")
        try:
            return [name_map[part] for part in parts]
        except KeyError:
            raise InvalidSettingsOrderError(self.__class__.__name__, param_name, order, name_map)

    def assert_no_duplicates_in_orders(self):
        """Checks if settings are not duplicated in after/before section and in the same section itself."""
        orders = {
            "keyword_before": set(self.keyword_before),
            "keyword_after": set(self.keyword_after),
            "test_before": set(self.test_before),
            "test_after": set(self.test_after),
        }
        # check if there is no duplicate in single order, ie test_after=setup,setup
        for name, order_set in orders.items():
            if len(self.__dict__[name]) != len(order_set):
                raise DuplicateInSettingsOrderError(self.__class__.__name__, name, self.__dict__[name])
        # check if there is no duplicate in opposite orders, ie test_before=tags test_after=tags
        shared_keyword = orders["keyword_before"].intersection(orders["keyword_after"])
        shared_test = orders["test_before"].intersection(orders["test_after"])
        if shared_keyword:
            raise SettingInBothOrdersError(self.__class__.__name__, "keyword_before", "keyword_after", shared_keyword)
        if shared_test:
            raise SettingInBothOrdersError(self.__class__.__name__, "test_before", "test_after", shared_test)

    @skip_section_if_disabled
    def visit_Section(self, node):  # noqa
        return self.generic_visit(node)

    @skip_if_disabled
    def visit_Keyword(self, node):  # noqa
        return self.order_settings(node, self.all_keyword_settings, self.keyword_before, self.keyword_after)

    @skip_if_disabled
    def visit_TestCase(self, node):  # noqa
        return self.order_settings(node, self.all_test_settings, self.test_before, self.test_after)

    def order_settings(self, node, setting_types, before, after):
        if not node.body:
            return node
        settings = dict()
        not_settings, trailing_after = [], []
        after_seen = False
        # when after_seen is set to True then all statements go to trailing_after and last non data
        # will be appended after tokens defined in `after` set (like [Return])
        comment = []
        for child in node.body:
            if isinstance(child, Comment) and child.lineno == node.lineno:
                comment.append(child)
            elif getattr(child, "type", "invalid") in setting_types:
                after_seen = after_seen or child.type in after
                settings[child.type] = child
            elif after_seen:
                trailing_after.append(child)
            else:
                not_settings.append(child)
        # comments after last data statement are considered as comment outside body
        trailing_non_data = []
        while trailing_after and isinstance(trailing_after[-1], (EmptyLine, Comment)):
            trailing_non_data.insert(0, trailing_after.pop())
        not_settings += trailing_after
        node.body = (
            comment
            + self.add_in_order(before, settings)
            + not_settings
            + self.add_in_order(after, settings)
            + trailing_non_data
        )
        return node

    @staticmethod
    def add_in_order(order, settings_in_node):
        return [settings_in_node[token_type] for token_type in order if token_type in settings_in_node]



from typing import List, Optional, Dict, Iterator, Tuple

from robocorp_ls_core.lsp import LocationTypedDict, RangeTypedDict, PositionTypedDict
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.protocols import (
    ICompletionContext,
    IRobotDocument,
    IKeywordFound,
    IVariablesCollector,
    IVariableFound,
    cast_to_keyword_definition,
    AbstractVariablesCollector,
    cast_to_variable_definition,
    VarTokenInfo,
    VariableKind,
    KeywordUsageInfo,
)
import typing
from robocorp_ls_core.protocols import check_implements
from robocorp_ls_core.basic import isinstance_name, normalize_filename


log = get_logger(__name__)


def matches_source(s1: str, s2: str) -> bool:
    if s1 == s2:
        return True

    return normalize_filename(s1) == normalize_filename(s2)


class _VariableDefinitionsCollector(AbstractVariablesCollector):
    def __init__(self, robot_string_matcher):
        from robotframework_ls.impl.string_matcher import RobotStringMatcher

        self.robot_string_matcher: RobotStringMatcher = robot_string_matcher
        self.matches: List[IVariableFound] = []

    def accepts(self, variable_name):
        return self.robot_string_matcher.is_variable_name_match(variable_name)

    def on_variable(self, variable_found: IVariableFound):
        self.matches.append(variable_found)

    def __typecheckself__(self) -> None:
        _: IVariablesCollector = check_implements(self)


def iter_variable_references_in_doc(
    completion_context: ICompletionContext,
    variable_found: IVariableFound,
    argument_var_references_computer: Optional[
        "_NamedArgumentVarReferencesComputer"
    ] = None,
) -> Iterator[RangeTypedDict]:
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.ast_utils import create_range_from_token
    from robotframework_ls.impl.string_matcher import RobotStringMatcher
    from robotframework_ls.impl.variable_completions import collect_variables
    from robotframework_ls.impl.text_utilities import normalize_robot_name
    from robotframework_ls.impl.variable_completions import collect_local_variables
    from robotframework_ls.impl.ast_utils import get_local_variable_stack_and_node

    normalized_name = normalize_robot_name(variable_found.variable_name)
    robot_string_matcher = RobotStringMatcher(normalized_name)

    # Collector for any variable with the same name.
    collector = _VariableDefinitionsCollector(robot_string_matcher)

    if argument_var_references_computer is None:
        argument_var_references_computer = _NamedArgumentVarReferencesComputer(
            completion_context, variable_found
        )

    if argument_var_references_computer.check_keyword_usage_normalized_name:
        lst = _PreventDuplicatesInList()
        argument_var_references_computer.add_references_to_named_keyword_arguments_from_doc(
            completion_context, lst
        )
        for entry in lst.lst:
            yield entry["range"]

    ast = completion_context.get_ast()
    if ast is not None:
        # Get references.
        var_token_info: VarTokenInfo
        if variable_found.is_local_variable:
            # For local variables we must have the stack
            stack = variable_found.stack
            assert stack

            # Just search the current stack.
            stack, stack_node = get_local_variable_stack_and_node(stack)

            for var_token_info in ast_utils.iter_variable_references(stack_node):
                completion_context.check_cancelled()

                if not robot_string_matcher.is_variable_name_match(
                    var_token_info.token.value
                ):
                    continue

                yield create_range_from_token(var_token_info.token)

            # Get definitions (only local).
            cp = completion_context.create_copy_with_selection(
                line=variable_found.lineno, col=variable_found.col_offset
            )
            token_info = cp.get_current_token()
            assert token_info
            collect_local_variables(cp, collector, token_info)

        else:
            # i.e.: For globals collect all globals as well as locals overriding
            # the global value.
            for var_token_info in ast_utils.iter_variable_references(ast):
                completion_context.check_cancelled()

                if not robot_string_matcher.is_variable_name_match(
                    var_token_info.token.value
                ):
                    continue

                yield create_range_from_token(var_token_info.token)

            # Get definitions (all).
            collect_variables(completion_context, collector, only_current_doc=True)

        variable: IVariableFound
        for variable in collector.matches:

            start: PositionTypedDict = {
                "line": variable.lineno,
                "character": variable.col_offset,
            }
            end: PositionTypedDict = {
                "line": variable.lineno,
                "character": variable.end_col_offset,
            }
            variable_range: RangeTypedDict = {"start": start, "end": end}
            yield variable_range


def iter_keyword_usage_references_in_doc(
    completion_context: ICompletionContext,
    doc: IRobotDocument,
    normalized_name: str,
    keyword_found: Optional[IKeywordFound],
) -> Iterator[Tuple[KeywordUsageInfo, bool, str, str]]:
    """
    :param keyword_found: if given, we'll match if the definition actually
    maps to the proper place (if not given, we'll just match based on the name
    without verifying if the definition is the same).
    """
    from robotframework_ls.impl import ast_utils
    from robotframework_ls.impl.find_definition import find_definition
    from robotframework_ls.impl.text_utilities import normalize_robot_name
    from robotframework_ls.impl.text_utilities import matches_name_with_variables

    ast = doc.get_ast()
    if ast is not None:
        has_var_in_name = "{" in normalized_name

        # Dict with normalized name -> whether it was found or not previously.
        found_in_this_doc: Dict[str, bool] = {}

        # Ok, we have the document, now, load the usages.
        for keyword_usage_info in ast_utils.iter_keyword_usage_tokens(
            ast, collect_args_as_keywords=True
        ):
            completion_context.check_cancelled()
            keword_name_possibly_dotted = keyword_usage_info.name
            found_dot_in_usage = "." in keword_name_possibly_dotted
            if found_dot_in_usage:
                keword_name_not_dotted = keword_name_possibly_dotted.split(".")[-1]
            else:
                keword_name_not_dotted = keword_name_possibly_dotted

            keword_name_not_dotted_normalized = normalize_robot_name(
                keword_name_not_dotted
            )
            if keword_name_not_dotted_normalized == normalized_name or (
                has_var_in_name
                and matches_name_with_variables(
                    keword_name_not_dotted_normalized, normalized_name
                )
            ):
                found_once_in_this_doc = found_in_this_doc.get(
                    keword_name_possibly_dotted
                )
                token = keyword_usage_info.token

                line = token.lineno - 1

                if keyword_found is not None:
                    if found_once_in_this_doc is None:
                        # Verify if it's actually the same one (not one defined in
                        # a different place with the same name).

                        new_ctx = completion_context.create_copy_doc_line_col(
                            doc, line, token.col_offset
                        )
                        definitions = find_definition(new_ctx)
                        for definition in definitions:
                            found = matches_source(
                                definition.source, keyword_found.source
                            )

                            if found:
                                found_once_in_this_doc = found_in_this_doc[
                                    keword_name_possibly_dotted
                                ] = True
                                break
                        else:
                            found_once_in_this_doc = found_in_this_doc[
                                keword_name_possibly_dotted
                            ] = False
                            continue

                    if not found_once_in_this_doc:
                        continue

                yield (
                    keyword_usage_info,
                    found_dot_in_usage,
                    keword_name_possibly_dotted,
                    keword_name_not_dotted,
                )


def iter_keyword_references_in_doc(
    completion_context: ICompletionContext,
    doc: IRobotDocument,
    normalized_name: str,
    keyword_found: Optional[IKeywordFound],
) -> Iterator[RangeTypedDict]:
    for (
        keyword_usage_info,
        found_dot_in_usage,
        keword_name_possibly_dotted,
        keword_name_not_dotted,
    ) in iter_keyword_usage_references_in_doc(
        completion_context, doc, normalized_name, keyword_found
    ):

        token = keyword_usage_info.token

        line = token.lineno - 1

        token = keyword_usage_info.token
        if found_dot_in_usage:
            # We need to create a new token because we just want to match the name part.
            col_offset = token.col_offset + (
                len(keword_name_possibly_dotted) - len(keword_name_not_dotted)
            )
            end_col_offset = token.end_col_offset
        else:
            col_offset = token.col_offset
            end_col_offset = token.end_col_offset

        # Ok, we found it, let's add it to the result.
        yield {
            "start": {
                "line": line,
                "character": col_offset,
            },
            "end": {
                "line": line,
                "character": end_col_offset,
            },
        }


def collect_variable_references(
    completion_context: ICompletionContext, var_token_info: VarTokenInfo
):
    from robotframework_ls.impl.find_definition import find_variable_definition

    var_definitions = find_variable_definition(completion_context, var_token_info)
    if not var_definitions:
        return []

    variable_found_lst: List[IVariableFound] = []

    for var_definition in var_definitions:
        as_variable_definition = cast_to_variable_definition(var_definition)
        if as_variable_definition:
            v = as_variable_definition.variable_found
            variable_found_lst.append(v)

    if not variable_found_lst:
        return []

    for v in variable_found_lst:
        if not v.is_local_variable:
            # I.e.: prefer globals (in which case we'll also collect shadowed
            # assigns in local scopes).
            variable_found = v
            break
    else:
        variable_found = next(iter(variable_found_lst))

    return _references_for_variable_found(completion_context, variable_found)


def references(
    completion_context: ICompletionContext, include_declaration: bool
) -> List[LocationTypedDict]:

    var_token_info = completion_context.get_current_variable()
    if var_token_info is not None:
        return collect_variable_references(completion_context, var_token_info)

    token_info = completion_context.get_current_token()
    if token_info is None:
        return []

    keyword_found: IKeywordFound
    if token_info.token.type == token_info.token.KEYWORD_NAME:
        if isinstance_name(token_info.node, "KeywordName"):
            from robotframework_ls.impl.find_definition import find_keyword_definition

            definitions = find_keyword_definition(completion_context, token_info)
            if definitions:
                for definition in definitions:
                    as_keyword_definition = cast_to_keyword_definition(definition)
                    if as_keyword_definition:
                        keyword_found = as_keyword_definition.keyword_found
                        return references_for_keyword_found(
                            completion_context, keyword_found, include_declaration
                        )

    current_keyword_definition_and_usage_info = (
        completion_context.get_current_keyword_definition_and_usage_info()
    )
    if current_keyword_definition_and_usage_info is not None:
        completion_context.monitor.check_cancelled()
        keyword_definition, _usage_info = current_keyword_definition_and_usage_info

        keyword_found = keyword_definition.keyword_found
        return references_for_keyword_found(
            completion_context, keyword_found, include_declaration
        )

    return []


class _NamedArgumentVarReferencesComputer:
    """
    A helper to handle the case where we also need to rename named arguments.

    To do this we need to:
    1. Get references to the keyword
    2. Check if any of its arguments has something as 'var_name=xxx'.
    3. Create the reference to the 'var_name'.
    """

    def __init__(
        self,
        initial_completion_context: ICompletionContext,
        variable_found: IVariableFound,
    ):
        from robotframework_ls.impl.ast_utils import get_local_variable_stack_and_node
        from robotframework_ls.impl.find_definition import find_keyword_definition
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        self.check_keyword_usage_keyword_found = None
        self.check_keyword_usage_normalized_name = None
        self.var_name_normalized = normalize_robot_name(variable_found.variable_name)

        if (
            variable_found.variable_kind == VariableKind.ARGUMENT
            and variable_found.stack
        ):
            _, keyword_or_test_case_node = get_local_variable_stack_and_node(
                variable_found.stack
            )
            if keyword_or_test_case_node.__class__.__name__ == "Keyword":
                cp = initial_completion_context.create_copy_with_selection(
                    keyword_or_test_case_node.lineno - 1,
                    keyword_or_test_case_node.col_offset,
                )
                cp_token_info = cp.get_current_token()
                if cp_token_info:
                    found = find_keyword_definition(
                        cp,
                        cp_token_info,
                    )

                    for keyword_found_definition in found or ():
                        self.check_keyword_usage_keyword_found = (
                            keyword_found_definition.keyword_found
                        )
                        self.check_keyword_usage_normalized_name = normalize_robot_name(
                            keyword_found_definition.keyword_name
                        )
                        break

    def add_references_to_named_keyword_arguments_from_doc(
        self,
        new_completion_context: ICompletionContext,
        ret: "_PreventDuplicatesInList",
    ):
        from robotframework_ls.impl.variable_resolve import find_split_index
        from robotframework_ls.impl.text_utilities import normalize_robot_name

        if not self.check_keyword_usage_normalized_name:
            return

        for (
            keyword_usage_info,
            _found_dot_in_usage,
            _keword_name_possibly_dotted,
            _keword_name_not_dotted,
        ) in iter_keyword_usage_references_in_doc(
            new_completion_context,
            new_completion_context.doc,
            self.check_keyword_usage_normalized_name,
            self.check_keyword_usage_keyword_found,
        ):
            for token in keyword_usage_info.node.tokens:
                if token.type == token.ARGUMENT:
                    split_eq = find_split_index(token.value)
                    if split_eq > 0:
                        arg_name = normalize_robot_name(token.value[:split_eq])
                        if arg_name == self.var_name_normalized:
                            start: PositionTypedDict = {
                                "line": token.lineno - 1,
                                "character": token.col_offset,
                            }
                            end: PositionTypedDict = {
                                "line": token.lineno - 1,
                                "character": token.col_offset + split_eq,
                            }
                            ref_range: RangeTypedDict = {"start": start, "end": end}
                            ret.append(
                                {
                                    "uri": new_completion_context.doc.uri,
                                    "range": ref_range,
                                }
                            )


class _PreventDuplicatesInList:
    def __init__(self):
        self.lst: List[LocationTypedDict] = []
        self._found = set()

    def append(self, location: LocationTypedDict):
        key = (
            location["uri"],
            location["range"]["start"]["line"],
            location["range"]["start"]["character"],
            location["range"]["end"]["line"],
            location["range"]["end"]["character"],
        )
        if key in self._found:
            return
        self._found.add(key)
        self.lst.append(location)


def _references_for_variable_found(
    initial_completion_context: ICompletionContext,
    variable_found: IVariableFound,
):
    from robotframework_ls.impl.text_utilities import normalize_robot_name

    ret = _PreventDuplicatesInList()

    is_local_variable = variable_found.is_local_variable

    from robotframework_ls.impl.workspace_symbols import iter_symbols_caches

    named_argument_var_references_computer = _NamedArgumentVarReferencesComputer(
        initial_completion_context, variable_found
    )

    # Initial doc (need to get local scope).
    ref_range: RangeTypedDict
    for ref_range in iter_variable_references_in_doc(
        initial_completion_context,
        variable_found,
        named_argument_var_references_computer,
    ):
        ret.append({"uri": initial_completion_context.doc.uri, "range": ref_range})

    if (
        is_local_variable
        and not named_argument_var_references_computer.check_keyword_usage_keyword_found
    ):
        return ret.lst

    normalized_variable_name = normalize_robot_name(variable_found.variable_name)

    for symbols_cache in iter_symbols_caches(
        None,
        initial_completion_context,
        force_all_docs_in_workspace=True,
        timeout=999999,
    ):
        initial_completion_context.check_cancelled()

        # If it's a local variable we may still need to search for named arguments...
        if (
            is_local_variable
            and named_argument_var_references_computer.check_keyword_usage_normalized_name
        ):
            if not symbols_cache.has_keyword_usage(
                named_argument_var_references_computer.check_keyword_usage_normalized_name
            ):
                continue

        elif not is_local_variable:
            if not symbols_cache.has_global_variable_definition(
                normalized_variable_name
            ) and not symbols_cache.has_variable_reference(normalized_variable_name):
                continue

        doc: Optional[IRobotDocument] = symbols_cache.get_doc()
        if doc is None:
            uri = symbols_cache.get_uri()
            if uri is None:
                continue

            doc = typing.cast(
                Optional[IRobotDocument],
                initial_completion_context.workspace.get_document(
                    doc_uri=uri, accept_from_file=True
                ),
            )

            if doc is None:
                log.debug(
                    "Unable to load document for getting references with uri: %s",
                    uri,
                )
                continue

        if initial_completion_context.doc.uri == doc.uri:
            continue  # Skip (already analyzed).

        new_completion_context = initial_completion_context.create_copy(doc)
        if not is_local_variable:
            # Collect references to global variables as well as named arguments.
            for ref_range in iter_variable_references_in_doc(
                new_completion_context,
                variable_found,
                named_argument_var_references_computer,
            ):
                ret.append({"uri": doc.uri, "range": ref_range})
        else:
            # We still need to collect references to named arguments.
            named_argument_var_references_computer.add_references_to_named_keyword_arguments_from_doc(
                new_completion_context, ret
            )

    return ret.lst


def references_for_keyword_found(
    completion_context: ICompletionContext,
    keyword_found: IKeywordFound,
    include_declaration: bool,
) -> list:
    from robocorp_ls_core import uris
    from robotframework_ls.impl.text_utilities import normalize_robot_name

    ret = _PreventDuplicatesInList()

    normalized_name = normalize_robot_name(keyword_found.keyword_name)
    # Ok, we have the keyword definition, now, we must actually look for the
    # references...
    if include_declaration:
        ret.append(
            {
                "uri": uris.from_fs_path(keyword_found.source),
                "range": {
                    "start": {
                        "line": keyword_found.lineno,
                        "character": keyword_found.col_offset,
                    },
                    "end": {
                        "line": keyword_found.end_lineno,
                        "character": keyword_found.end_col_offset,
                    },
                },
            }
        )

    from robotframework_ls.impl.workspace_symbols import iter_symbols_caches

    for symbols_cache in iter_symbols_caches(
        None, completion_context, force_all_docs_in_workspace=True, timeout=999999
    ):
        completion_context.check_cancelled()
        if symbols_cache.has_keyword_usage(normalized_name):
            doc: Optional[IRobotDocument] = symbols_cache.get_doc()
            if doc is None:
                uri = symbols_cache.get_uri()
                if uri is None:
                    continue

                doc = typing.cast(
                    Optional[IRobotDocument],
                    completion_context.workspace.get_document(
                        doc_uri=uri, accept_from_file=True
                    ),
                )

                if doc is None:
                    log.debug(
                        "Unable to load document for getting references with uri: %s",
                        uri,
                    )
                    continue

            ref_range: RangeTypedDict
            cp = completion_context.create_copy(doc)
            for ref_range in iter_keyword_references_in_doc(
                cp, doc, normalized_name, keyword_found
            ):
                ret.append({"uri": doc.uri, "range": ref_range})

    return ret.lst



from robot.api.parsing import Token

from robotidy.disablers import skip_if_disabled
from robotidy.exceptions import InvalidParameterValueError
from robotidy.skip import Skip
from robotidy.transformers import Transformer
from robotidy.transformers.run_keywords import get_run_keywords
from robotidy.utils import (
    collect_comments_from_tokens,
    get_new_line,
    is_token_value_in_tokens,
    join_tokens_with_token,
    merge_comments_into_one,
    normalize_name,
    split_on_token_type,
    split_on_token_value,
)


class IndentNestedKeywords(Transformer):
    """
    Format indentation inside run keywords variants such as ``Run Keywords`` or
    ``Run Keyword And Continue On Failure``.

    Keywords inside run keywords variants are detected and
    whitespace is formatted to outline them. This code:

    ```robotframework
        Run Keyword    Run Keyword If    ${True}    Run keywords   Log    foo    AND    Log    bar    ELSE    Log    baz
    ```

    will be transformed to:

    ```robotframework
        Run Keyword
        ...    Run Keyword If    ${True}
        ...        Run keywords
        ...            Log    foo
        ...            AND
        ...            Log    bar
        ...    ELSE
        ...        Log    baz
    ```

    ``AND`` argument inside ``Run Keywords`` can be handled in different ways. It is controlled via ``indent_and``
    parameter. For more details see the full documentation.

    To skip formatting run keywords inside settings (such as ``Suite Setup``, ``[Setup]``, ``[Teardown]`` etc.) set
    ``skip_settings`` to ``True``.
    """

    ENABLED = False
    HANDLES_SKIP = frozenset({"skip_settings"})

    def __init__(self, indent_and: str = "split", skip: Skip = None):
        super().__init__(skip=skip)
        self.indent_and = indent_and
        self.validate_indent_and()
        self.run_keywords = get_run_keywords()

    def validate_indent_and(self):
        modes = {"keep_in_line", "split", "split_and_indent"}
        if self.indent_and not in modes:
            raise InvalidParameterValueError(
                self.__class__.__name__,
                "indent_and",
                self.indent_and,
                f"Select one of: {','.join(modes)}",
            )

    def get_run_keyword(self, kw_name):
        kw_norm = normalize_name(kw_name)
        return self.run_keywords.get(kw_norm, None)

    def get_setting_lines(self, node):  # noqa
        if self.skip.setting("any") or node.errors or not len(node.data_tokens) > 1:
            return None
        run_keyword = self.get_run_keyword(node.data_tokens[1].value)
        if not run_keyword:
            return None
        return self.parse_sub_kw(node.data_tokens[1:])

    def get_separator(self, column=1, continuation=False):
        if continuation:
            separator = self.formatting_config.continuation_indent * column
        else:
            separator = self.formatting_config.separator * column
        return Token(Token.SEPARATOR, separator)

    def parse_keyword_lines(self, lines, tokens, new_line, eol):
        separator = self.get_separator()
        for column, line in lines[1:]:
            tokens.extend(new_line)
            tokens.append(self.get_separator(column, True))
            tokens.extend(join_tokens_with_token(line, separator))
        tokens.append(eol)
        return tokens

    @skip_if_disabled
    def visit_SuiteSetup(self, node):  # noqa
        lines = self.get_setting_lines(node)
        if not lines:
            return node
        comments = collect_comments_from_tokens(node.tokens, indent=None)
        separator = self.get_separator()
        new_line = get_new_line()
        tokens = [node.data_tokens[0], separator, *join_tokens_with_token(lines[0][1], separator)]
        node.tokens = self.parse_keyword_lines(lines, tokens, new_line, eol=node.tokens[-1])
        return (*comments, node)

    visit_SuiteTeardown = visit_TestSetup = visit_TestTeardown = visit_SuiteSetup

    @skip_if_disabled
    def visit_Setup(self, node):  # noqa
        lines = self.get_setting_lines(node)
        if not lines:
            return node
        indent = node.tokens[0]
        separator = self.get_separator()
        new_line = get_new_line(indent)
        tokens = [indent, node.data_tokens[0], separator, *join_tokens_with_token(lines[0][1], separator)]
        comment = merge_comments_into_one(node.tokens)
        if comment:
            # need to add comments on first line for [Setup] / [Teardown] settings
            comment_sep = Token(Token.SEPARATOR, "  ")
            tokens.extend([comment_sep, Token(Token.COMMENT, comment)])
        node.tokens = self.parse_keyword_lines(lines, tokens, new_line, eol=node.tokens[-1])
        return node

    visit_Teardown = visit_Setup

    @skip_if_disabled
    def visit_KeywordCall(self, node):  # noqa
        if node.errors or not node.keyword:
            return node
        run_keyword = self.get_run_keyword(node.keyword)
        if not run_keyword:
            return node

        indent = node.tokens[0]
        comments = collect_comments_from_tokens(node.tokens, indent)
        assign, kw_tokens = split_on_token_type(node.data_tokens, Token.KEYWORD)
        lines = self.parse_sub_kw(kw_tokens)
        if not lines:
            return node

        separator = self.get_separator()
        tokens = [indent]
        if assign:
            tokens.extend([*join_tokens_with_token(assign, separator), separator])
        tokens.extend(join_tokens_with_token(lines[0][1], separator))
        new_line = get_new_line(indent)
        node.tokens = self.parse_keyword_lines(lines, tokens, new_line, eol=node.tokens[-1])
        return (*comments, node)

    def parse_sub_kw(self, tokens, column=0):
        if not tokens:
            return []
        run_keyword = self.get_run_keyword(tokens[0].value)
        if not run_keyword:
            return [(column, list(tokens))]
        lines = [(column, tokens[: run_keyword.resolve])]
        tokens = tokens[run_keyword.resolve :]
        if run_keyword.branches:
            if "ELSE IF" in run_keyword.branches:
                while is_token_value_in_tokens("ELSE IF", tokens):
                    column = max(column, 1)
                    prefix, branch, tokens = split_on_token_value(tokens, "ELSE IF", 2)
                    lines.extend(self.parse_sub_kw(prefix, column + 1))
                    lines.append((column, branch))
            if "ELSE" in run_keyword.branches and is_token_value_in_tokens("ELSE", tokens):
                return self.split_on_else(tokens, lines, column)
        elif run_keyword.split_on_and:
            return self.split_on_and(tokens, lines, column)
        return lines + self.parse_sub_kw(tokens, column + 1)

    def split_on_else(self, tokens, lines, column):
        column = max(column, 1)
        prefix, branch, tokens = split_on_token_value(tokens, "ELSE", 1)
        lines.extend(self.parse_sub_kw(prefix, column + 1))
        lines.append((column, branch))
        lines.extend(self.parse_sub_kw(tokens, column + 1))
        return lines

    def split_on_and(self, tokens, lines, column):
        if is_token_value_in_tokens("AND", tokens):
            while is_token_value_in_tokens("AND", tokens):
                prefix, branch, tokens = split_on_token_value(tokens, "AND", 1)
                if self.indent_and == "keep_in_line":
                    lines.extend(self.parse_sub_kw(prefix + branch, column + 1))
                else:
                    indent = int(self.indent_and == "split_and_indent")  # indent = 1 for split_and_indent, else 0
                    lines.extend(self.parse_sub_kw(prefix, column + 1 + indent))
                    lines.append((column + 1, branch))
            indent = int(self.indent_and == "split_and_indent")  # indent = 1 for split_and_indent, else 0
            lines.extend(self.parse_sub_kw(tokens, column + 1 + indent))
        else:
            lines.extend([(column + 1, [kw_token]) for kw_token in tokens])
        return lines



import functools
import re

from robot.api.parsing import Comment, ModelVisitor, Token


def skip_if_disabled(func):
    """
    Do not transform node if it's not within passed ``start_line`` and ``end_line`` or
    it does match any ``# robotidy: off`` disabler
    """

    @functools.wraps(func)
    def wrapper(self, node, *args, **kwargs):
        if self.disablers.is_node_disabled(node):
            return node
        return func(self, node, *args, **kwargs)

    return wrapper


def skip_section_if_disabled(func):
    """
    Does the same checks as ``skip_if_disabled`` and additionally checks
    if the section header does not contain disabler
    """

    @functools.wraps(func)
    def wrapper(self, node, *args, **kwargs):
        if self.disablers.is_node_disabled(node):
            return node
        if self.disablers.is_header_disabled(node.lineno):
            return node
        return func(self, node, *args, **kwargs)

    return wrapper


def is_line_start(node):
    for token in node.tokens:
        if token.type == Token.SEPARATOR:
            continue
        return token.col_offset == 0
    return False


class DisabledLines:
    def __init__(self, start_line, end_line, file_end):
        self.start_line = start_line
        self.end_line = end_line
        self.file_end = file_end
        self.lines = []
        self.disabled_headers = set()

    def add_disabler(self, start_line, end_line):
        self.lines.append((start_line, end_line))

    def add_disabled_header(self, lineno):
        self.disabled_headers.add(lineno)

    def parse_global_disablers(self):
        if not self.start_line:
            return
        end_line = self.end_line if self.end_line else self.start_line
        if self.start_line > 1:
            self.add_disabler(1, self.start_line - 1)
        if end_line < self.file_end:
            self.add_disabler(end_line + 1, self.file_end)

    def sort_disablers(self):
        self.lines = sorted(self.lines, key=lambda x: x[0])

    def is_header_disabled(self, line):
        return line in self.disabled_headers

    def is_node_disabled(self, node, full_match=True):
        if full_match:
            for start_line, end_line in self.lines:
                # lines are sorted on start_line, so we can return on first match
                if end_line >= node.end_lineno:
                    return start_line <= node.lineno
        else:
            for start_line, end_line in self.lines:
                if node.lineno <= end_line and node.end_lineno >= start_line:
                    return True
        return False


class RegisterDisablers(ModelVisitor):
    def __init__(self, start_line, end_line):
        self.start_line = start_line
        self.end_line = end_line
        self.disablers = DisabledLines(self.start_line, self.end_line, None)
        self.disabler_pattern = re.compile(r"\s*#\s?robotidy:\s?(?P<disabler>on|off)")
        self.stack = []
        self.file_disabled = False

    def any_disabler_open(self):
        return any(disabler for disabler in self.stack)

    def get_disabler(self, comment):
        if not comment.value:
            return None
        return self.disabler_pattern.match(comment.value)

    def close_disabler(self, end_line):
        disabler = self.stack.pop()
        if disabler:
            self.disablers.add_disabler(disabler, end_line)

    def visit_File(self, node):  # noqa
        self.disablers = DisabledLines(self.start_line, self.end_line, node.end_lineno)
        self.disablers.parse_global_disablers()
        self.stack = []
        self.file_disabled = False
        self.generic_visit(node)
        self.disablers.sort_disablers()

    def visit_SectionHeader(self, node):  # noqa
        for comment in node.get_tokens(Token.COMMENT):
            disabler = self.get_disabler(comment)
            if disabler and disabler.group("disabler") == "off":
                self.disablers.add_disabled_header(node.lineno)
                break
        return self.generic_visit(node)

    def visit_TestCase(self, node):  # noqa
        self.stack.append(0)
        self.generic_visit(node)
        if self.file_disabled:  # stop visiting if whole file is disabled
            return
        self.close_disabler(node.end_lineno)

    def visit_Try(self, node):  # noqa
        self.generic_visit(node.header)
        self.stack.append(0)
        for statement in node.body:
            self.visit(statement)
        if self.file_disabled:  # stop visiting if whole file is disabled
            return
        self.close_disabler(node.end_lineno)
        tail = node
        while tail.next:
            self.generic_visit(tail.header)
            self.stack.append(0)
            for statement in tail.body:
                self.visit(statement)
            if self.file_disabled:  # stop visiting if whole file is disabled
                return
            end_line = tail.next.lineno - 1 if tail.next else tail.end_lineno
            self.close_disabler(end_line=end_line)
            tail = tail.next

    visit_Keyword = visit_Section = visit_For = visit_ForLoop = visit_If = visit_While = visit_TestCase

    def visit_Statement(self, node):  # noqa
        if isinstance(node, Comment):
            comment = node.get_token(Token.COMMENT)
            disabler = self.get_disabler(comment)
            if not disabler:
                return
            index = 0 if is_line_start(node) else -1
            if disabler.group("disabler") == "on":
                if not self.stack[index]:  # no disabler open
                    return
                self.disablers.add_disabler(self.stack[index], node.lineno)
                self.stack[index] = 0
            else:
                if node.lineno == 1 and index == 0:
                    self.file_disabled = True
                elif not self.stack[index]:
                    self.stack[index] = node.lineno
        else:
            # inline disabler
            if self.any_disabler_open():
                return
            for comment in node.get_tokens(Token.COMMENT):
                disabler = self.get_disabler(comment)
                if disabler and disabler.group("disabler") == "off":
                    self.disablers.add_disabler(node.lineno, node.end_lineno)



"""
Transformers are classes used to transform passed Robot Framework code model.

To create your own transformer you need to create file with the same name as your transformer class. Your class
need to inherit from ``ModelTransformer`` or ``ast.NodeTransformer`` class. Finally put name of your transformer in
``TRANSFORMERS`` variable in this file.

If you don't want to run your transformer by default and only when calling robotidy with --transform YourTransformer
then add ``ENABLED = False`` class attribute inside.
"""
import copy
import textwrap
from itertools import chain
from typing import Dict, Optional

try:
    import rich_click as click
except ImportError:
    import click

from robot.api.parsing import ModelTransformer
from robot.errors import DataError
from robot.utils.importer import Importer

from robotidy.exceptions import ImportTransformerError, InvalidParameterError, InvalidParameterFormatError
from robotidy.skip import Skip, SkipConfig
from robotidy.utils import ROBOT_VERSION, RecommendationFinder

TRANSFORMERS = [
    "AddMissingEnd",
    "NormalizeSeparators",
    "DiscardEmptySections",
    "MergeAndOrderSections",
    "RemoveEmptySettings",
    "NormalizeAssignments",
    "OrderSettings",
    "OrderSettingsSection",
    "NormalizeTags",
    "OrderTags",
    "IndentNestedKeywords",
    "AlignSettingsSection",
    "AlignVariablesSection",
    "AlignTemplatedTestCases",
    "AlignTestCasesSection",
    "AlignKeywordsSection",
    "NormalizeNewLines",
    "NormalizeSectionHeaderName",
    "NormalizeSettingName",
    "ReplaceRunKeywordIf",
    "SplitTooLongLine",
    "SmartSortKeywords",
    "RenameTestCases",
    "RenameKeywords",
    "ReplaceReturns",
    "ReplaceBreakContinue",
    "InlineIf",
]


IMPORTER = Importer()


class TransformerParameter:
    def __init__(self, name, default_value):
        self.name = name
        self.value = default_value

    def __str__(self):
        if self.value is not None and str(self.value) != "":
            return f"{self.name} : {self.value}"
        return self.name


class TransformerContainer:
    """
    Stub for transformer container class that holds the transformer instance and its metadata.
    """

    def __init__(self, instance, argument_names, spec):
        self.instance = instance
        self.name = instance.__class__.__name__
        self.enabled_by_default = getattr(instance, "ENABLED", True)
        self.parameters = self.get_parameters(argument_names, spec)

    def get_parameters(self, argument_names, spec):
        params = []
        for arg in argument_names:
            if arg == "enabled":
                default = self.enabled_by_default
            else:
                default = spec.defaults.get(arg, None)
            params.append(TransformerParameter(arg, default))
        return params

    def __str__(self):
        s = f"## Transformer {self.name}\n" + textwrap.dedent(self.instance.__doc__)
        if self.parameters:
            s += "\nSupported parameters:\n  - " + "\n - ".join(str(param) for param in self.parameters) + "\n"
        s += f"\nSee <https://robotidy.readthedocs.io/en/latest/transformers/{self.name}.html> for more examples."
        return s


class Transformer(ModelTransformer):
    def __init__(self, skip: Optional[Skip] = None):
        self.formatting_config = None  # to make lint happy (we're injecting the configs)
        self.transformers: Dict = dict()
        self.disablers = None
        self.skip = skip


def import_transformer(name, args, skip):
    short_name = name.split(".")[-1]
    try:
        imported_class = IMPORTER.import_class_or_module(name)
        spec = IMPORTER._get_arg_spec(imported_class)
        handles_skip = getattr(imported_class, "HANDLES_SKIP", {})
        positional, named, argument_names = resolve_args(short_name, spec, args, skip, handles_skip=handles_skip)
    except DataError:
        similar_finder = RecommendationFinder()
        similar = similar_finder.find_similar(short_name, TRANSFORMERS)
        raise ImportTransformerError(
            f"Importing transformer '{short_name}' failed. "
            f"Verify if correct name or configuration was provided.{similar}"
        ) from None
    instance = imported_class(*positional, **named)
    return TransformerContainer(instance, argument_names, spec)


def split_args_to_class_and_skip(args):
    filtered_args = []
    skip_args = {}
    for arg, value in args.items():
        if arg == "enabled":
            continue
        if arg in SkipConfig.HANDLES:
            skip_args[arg.replace("skip_", "")] = value
        else:
            filtered_args.append(f"{arg}={value}")
    return filtered_args, skip_args


def resolve_argument_names(argument_names, handles_skip):
    """Get transformer argument names with resolved skip parameters."""
    new_args = ["enabled"]
    if "skip" not in argument_names:
        return new_args + argument_names
    new_args.extend([arg for arg in argument_names if arg != "skip"])
    new_args.extend(arg for arg in sorted(handles_skip) if arg not in new_args)
    return new_args


def assert_handled_arguments(transformer, args, argument_names):
    """Check if provided arguments are handled by given transformer.
    Raises InvalidParameterError if arguments does not match."""
    arg_names = [arg.split("=")[0] for arg in args]
    for arg in arg_names:
        # it's fine to only check for first non-matching parameter
        if arg not in argument_names:
            similar_finder = RecommendationFinder()
            similar = similar_finder.find_similar(arg, argument_names)
            if not similar and argument_names:
                arg_names = "\n    " + "\n    ".join(argument_names)
                similar = f" This transformer accepts following arguments:{arg_names}"
            raise InvalidParameterError(transformer, similar) from None


def get_skip_args_from_spec(spec):
    """
    It is possible to override default skip value (such as skip_documentation
    from False to True in AlignKeywordsSection).
    This method iterate over spec and finds such overrides.
    """
    defaults = dict()
    for arg, value in spec.defaults.items():
        if arg in SkipConfig.HANDLES:
            defaults[arg.replace("skip_", "")] = value
    return defaults


def get_skip_class(spec, skip_args, global_skip):
    defaults = get_skip_args_from_spec(spec)
    defaults.update(skip_args)
    if global_skip is None:
        skip_config = SkipConfig()
    else:
        skip_config = copy.deepcopy(global_skip)
    skip_config.update_with_str_config(**defaults)
    return Skip(skip_config)


def resolve_args(transformer, spec, args, global_skip, handles_skip):
    """
    Use class definition to identify which arguments from configuration
    should be used to invoke it.

    First we're splitting arguments into class arguments and skip arguments
    (those that are handled by Skip class).
    Class arguments are resolved with their definition and if class accepts
    "skip" parameter the Skip class will be also added to class arguments.
    """
    args, skip_args = split_args_to_class_and_skip(args)
    argument_names = resolve_argument_names(spec.argument_names, handles_skip)
    assert_handled_arguments(transformer, args, argument_names)
    try:
        positional, named = spec.resolve(args)
        named = dict(named)
        if "skip" in spec.argument_names:
            named["skip"] = get_skip_class(spec, skip_args, global_skip)
        return positional, named, argument_names
    except ValueError as err:
        raise InvalidParameterError(transformer, f" {err}") from None


def resolve_core_import_path(name):
    """Append import path if transformer is core Robotidy transformer."""
    return f"robotidy.transformers.{name}" if name in TRANSFORMERS else name


def load_transformer(name, args, skip) -> Optional[TransformerContainer]:
    if not args.get("enabled", True):
        return None
    import_name = resolve_core_import_path(name)
    return import_transformer(import_name, args, skip)


def join_configs(args, config):
    # args are from --transform Name:param=value and config is from --configure
    temp_args = {}
    for arg in chain(args, config):
        param, value = arg.split("=", maxsplit=1)
        if param == "enabled":
            temp_args[param] = value.lower() == "true"
        else:
            temp_args[param] = value
    return temp_args


def get_args(transformer, allowed_mapped, config):
    try:
        return join_configs(allowed_mapped.get(transformer, ()), config.get(transformer, ()))
    except ValueError:
        raise InvalidParameterFormatError(transformer) from None


def validate_config(config, allowed_mapped):
    for transformer in config:
        if transformer in allowed_mapped or transformer in TRANSFORMERS:
            continue
        similar_finder = RecommendationFinder()
        similar = similar_finder.find_similar(transformer, TRANSFORMERS + list(allowed_mapped.keys()))
        raise ImportTransformerError(
            f"Configuring transformer '{transformer}' failed. " f"Verify if correct name was provided.{similar}"
        ) from None


def can_run_in_robot_version(transformer, overwritten, target_version):
    if not hasattr(transformer, "MIN_VERSION"):
        return True
    if target_version >= transformer.MIN_VERSION:
        return True
    if overwritten:
        # --transform TransformerDisabledInVersion or --configure TransformerDisabledInVersion:enabled=True
        if target_version == ROBOT_VERSION.major:
            click.echo(
                f"{transformer.__class__.__name__} transformer requires Robot Framework {transformer.MIN_VERSION}.* "
                f"version but you have {ROBOT_VERSION} installed. "
                f"Upgrade installed Robot Framework if you want to use this transformer."
            )
        else:
            click.echo(
                f"{transformer.__class__.__name__} transformer requires Robot Framework {transformer.MIN_VERSION}.* "
                f"version but you set --target-version rf{target_version}. "
                f"Set --target-version to rf{transformer.MIN_VERSION} or do not forcefully enable this transformer "
                f"with --transform / enable parameter."
            )
    return False


def assert_not_duplicated_transform(allowed_mapped, allowed_transformers):
    """
    Warns if user provides duplicated transformer names in --transform option.
    """
    if not allowed_mapped or len(allowed_mapped) == len(allowed_transformers):
        return
    transformers, reported = set(), set()
    for name, _ in allowed_transformers:
        if name in transformers and name not in reported:
            reported.add(name)
            click.echo(
                f"Duplicated transformer '{name}' in the transform option. "
                f"It will be run only once with the configuration from the last transform."
            )
        else:
            transformers.add(name)


def load_transformers(
    allowed_transformers,
    config,
    target_version,
    skip=None,
    allow_disabled=False,
    force_order=False,
    allow_version_mismatch=True,
):
    """Dynamically load all classes from this file with attribute `name` defined in allowed_transformers"""
    loaded_transformers = []
    allowed_mapped = {name: args for name, args in allowed_transformers} if allowed_transformers else {}
    assert_not_duplicated_transform(allowed_mapped, allowed_transformers)
    validate_config(config, allowed_mapped)
    if not force_order:
        for name in TRANSFORMERS:
            if not allowed_mapped or name in allowed_mapped:
                args = get_args(name, allowed_mapped, config)
                container = load_transformer(name, args, skip)
                if container is None:
                    continue
                enabled = getattr(container.instance, "ENABLED", True) or args.get("enabled", False)
                if allowed_mapped or allow_disabled or enabled:
                    overwritten = name in allowed_mapped or args.get("enabled", False)
                    if can_run_in_robot_version(
                        container.instance, overwritten=overwritten, target_version=target_version
                    ):
                        loaded_transformers.append(container)
                    elif allow_version_mismatch and allow_disabled:
                        setattr(container.instance, "ENABLED", False)
                        container.enabled_by_default = False
                        loaded_transformers.append(container)
    for name in allowed_mapped:
        if force_order or name not in TRANSFORMERS:
            args = get_args(name, allowed_mapped, config)
            container = load_transformer(name, args, skip)
            if container is not None:
                if can_run_in_robot_version(container.instance, overwritten=True, target_version=target_version):
                    loaded_transformers.append(container)
    return loaded_transformers
